import numpy as np
import photutils

from paos import logger
from paos.classes.zernike import Zernike


class WFO:
    """
    Physical optics wavefront propagation.
    Implements the paraxial theory described in
    `Lawrence et al., Applied Optics and Optical Engineering, Volume XI (1992) <https://ui.adsabs.harvard.edu/abs/1992aooe...11..125L>`_

    All units are meters.

    Parameters
    ----------
    beam_diameter: scalar
        the input beam diameter. Note that the input beam is always circular, regardless of
        whatever non-circular apodization the input pupil might apply.
    wl: scalar
        the wavelength
    grid_size: scalar
        grid size must be a power of 2
    zoom: scalar
        linear scaling factor of input beam.

    Attributes
    ----------
    wl: scalar
        the wavelength
    z: scalar
        current beam position along the z-axis (propagation axis).
        Initial value is 0
    w0: scalar
        pilot Gaussian beam waist.
        Initial value is beam_diameter/2
    zw0: scalar
        z-coordinate of the Gaussian beam waist.
        initial value is 0
    zr: scalar
        Rayleigh distance: :math:`\\pi w_{0}^{2} / \\lambda`
    rayleigh_factor: scalar
        Scale factor multiplying zr to determine 'I' and 'O' regions.
        Built in value is 2
    dx: scalar
        pixel sampling interval along x-axis
    dy: scalar
        pixel sampling interval along y-axis
    C: scalar
        curvature of the reference surface at beam position
    fratio: scalar
        pilot Gaussian beam f-ratio
    wfo: array [gridsize, gridsize], complex128
        the wavefront complex array
    amplitude: array [gridsize, gridsize], float64
        the wavefront amplitude array
    phase: array [gridsize, gridsize], float64
        the wavefront phase array in radians
    wz: scalar
        the Gaussian beam waist w(z) at current beam position
    distancetofocus: scalar
        the distance to focus from current beam position
    extent: tuple
        the physical coordinates of the wavefront bounding box (xmin, xmax, ymin, ymax).
        Can be used directly in im.set_extent.

    Returns
    -------
    out: an instance of wfo

    Example
    -------
    >>> import paos
    >>> import matplotlib.pyplot as plt
    >>> beam_diameter = 1.0  # m
    >>> wavelength = 3.0  # micron
    >>> grid_size = 512
    >>> zoom = 4
    >>> xdec, ydec = 0.0, 0.0
    >>> fig, (ax0, ax1) = plt.subplots(nrows=1, ncols=2, figsize=(12, 6))
    >>> wfo = paos.WFO(beam_diameter, 1.0e-6 * wavelength, grid_size, zoom)
    >>> wfo.aperture(xc=xdec, yc=ydec, r=beam_diameter/2, shape='circular')
    >>> wfo.make_stop()
    >>> ax0.imshow(wfo.amplitude)
    >>> wfo.lens(lens_fl=1.0)
    >>> wfo.propagate(dz=1.0)
    >>> ax1.imshow(wfo.amplitude)
    >>> plt.show()

    """

    def __init__(self, beam_diameter, wl, grid_size, zoom):

        assert np.log2(grid_size).is_integer(), "Grid size should be 2**n"
        assert zoom > 0, "zoom factor should be positive"
        assert beam_diameter > 0, "beam diameter should be positive"
        assert wl > 0, "a wavelength should be positive"

        self._wl = wl
        self._z = 0.0  # current beam z coordinate
        self._w0 = beam_diameter / 2.0  # beam waist
        self._zw0 = 0.0  # beam waist z coordinate
        self._zr = np.pi * self.w0**2 / wl  # Rayleigh distance

        self._rayleigh_factor = 2.0
        self._dx = beam_diameter * zoom / grid_size  # pixel size
        self._dy = beam_diameter * zoom / grid_size  # pixel size
        self._C = 0.0  # beam curvature, start with a planar wf
        self._fratio = np.inf  # Gaussian beam f-ratio

        grid_size = np.uint(grid_size)
        self._wfo = np.ones((grid_size, grid_size), dtype=np.complex128)

    @property
    def wl(self):
        return self._wl

    @property
    def z(self):
        return self._z

    @property
    def w0(self):
        return self._w0

    @property
    def zw0(self):
        return self._zw0

    @property
    def zr(self):
        return self._zr

    @property
    def rayleigh_factor(self):
        return self._rayleigh_factor

    @property
    def dx(self):
        return self._dx

    @property
    def dy(self):
        return self._dy

    @property
    def C(self):
        return self._C

    @property
    def fratio(self):
        return self._fratio

    @property
    def wfo(self):
        return self._wfo.copy()

    @property
    def amplitude(self):
        return np.abs(self._wfo)

    @property
    def phase(self):
        return np.angle(self._wfo)

    @property
    def wz(self):
        return self.w0 * np.sqrt(1.0 + ((self.z - self.zw0) / self.zr) ** 2)

    @property
    def distancetofocus(self):
        return self.zw0 - self.z

    @property
    def extent(self):
        return (
            -self._wfo.shape[1] // 2 * self.dx,
            (self._wfo.shape[1] // 2 - 1) * self.dx,
            -self._wfo.shape[0] // 2 * self.dy,
            (self._wfo.shape[0] // 2 - 1) * self.dy,
        )

    def make_stop(self):
        """
        Make current surface a stop.
        Stop here just means that the wf at current position is normalised to unit energy.
        """
        norm2 = np.sum(np.abs(self._wfo) ** 2)
        self._wfo /= np.sqrt(norm2)

    def aperture(
        self,
        xc,
        yc,
        hx=None,
        hy=None,
        r=None,
        shape="elliptical",
        tilt=None,
        obscuration=False,
    ):
        """
        Apply aperture mask

        Parameters
        ----------
        xc: scalar
            x-centre of the aperture
        yc: scalar
            y-centre of the aperture
        hx, hy: scalars
            semi-axes of shape 'elliptical' aperture, or full dimension of shape 'rectangular' aperture
        r: scalar
            radius of shape 'circular' aperture
        shape: string
            defines aperture shape. Can be 'elliptical', 'circular', 'rectangular'
        tilt: scalar
            tilt angle in degrees. Applies to shapes 'elliptical' and 'rectangular'.
        obscuration: boolean
            if True, aperture mask is converted into obscuration mask.

        """

        ixc = xc / self.dx + self._wfo.shape[1] / 2
        iyc = yc / self.dy + self._wfo.shape[0] / 2

        if shape == "elliptical":
            if hx is None or hy is None:
                logger.error("Semi major/minor axes not defined")
                raise AssertionError("Semi major/minor axes not defined")
            ihx = hx / self.dx
            ihy = hy / self.dy
            theta = 0.0 if tilt is None else np.deg2rad(tilt)
            aperture = photutils.aperture.EllipticalAperture(
                (ixc, iyc), ihx, ihy, theta=theta
            )
            mask = aperture.to_mask(method="exact").to_image(self._wfo.shape)
        elif shape == "circular":
            if r is None:
                logger.error("Radius not defined")
                raise AssertionError("Radius not defined")
            ihx = r / self.dx
            ihy = r / self.dy
            theta = 0.0
            aperture = photutils.aperture.EllipticalAperture(
                (ixc, iyc), ihx, ihy, theta=theta
            )
            mask = aperture.to_mask(method="exact").to_image(self._wfo.shape)
        elif shape == "rectangular":
            if hx is None or hy is None:
                logger.error("Semi major/minor axes not defined")
                raise AssertionError("Semi major/minor axes not defined")
            ihx = hx / self.dx
            ihy = hy / self.dy
            theta = 0.0 if tilt is None else np.deg2rad(tilt)
            aperture = photutils.aperture.RectangularAperture(
                (ixc, iyc), ihx, ihy, theta=theta
            )
            # Exact method not implemented in photutils 1.0.2
            mask = aperture.to_mask(method="subpixel", subpixels=32).to_image(
                self._wfo.shape
            )
        else:
            logger.error("Aperture {:s} not defined yet.".format(shape))
            raise ValueError("Aperture {:s} not defined yet.".format(shape))

        if obscuration:
            self._wfo *= 1 - mask
        else:
            self._wfo *= mask

        return aperture

    def insideout(self, z=None):
        """
        Check if z position is within the Rayleigh distance

        Parameters
        ----------
        z: scalar
            beam coordinate long propagation axis

        Returns
        -------
            out: string
                'I' if :math:`|z - z_{w0}| < z_{r}` else 'O'
        """
        if z is None:
            delta_z = self.z - self.zw0
        else:
            delta_z = z - self.zw0

        if np.abs(delta_z) < self.rayleigh_factor * self.zr:
            return "I"
        else:
            return "O"

    def lens(self, lens_fl):
        """
        Apply wavefront phase from paraxial lens

        Parameters
        ----------
        lens_fl: scalar
            Lens focal length. Positive for converging lenses. Negative for diverging lenses.

        Note
        ----------
            A paraxial lens imposes a quadratic phase shift.
        """

        wz = self.w0 * np.sqrt(1.0 + ((self.z - self.zw0) / self.zr) ** 2)
        delta_z = self.z - self.zw0

        propagator = self.insideout()

        # estimate Gaussian beam curvature after lens
        gCobj = delta_z / (
            delta_z**2 + self.zr**2
        )  # Gaussian beam curvature before lens
        gCima = gCobj - 1.0 / lens_fl  # Gaussian beam curvature after lens

        # update Gaussian beam parameters
        self._w0 = wz / np.sqrt(1.0 + (np.pi * wz**2 * gCima / self.wl) ** 2)
        self._zw0 = (
            -gCima / (gCima**2 + (self.wl / (np.pi * wz**2)) ** 2) + self.z
        )
        self._zr = np.pi * self.w0**2 / self.wl

        propagator = propagator + self.insideout()

        if propagator[0] == "I" or self.C == 0.0:
            Cobj = 0.0
        else:
            Cobj = 1 / delta_z

        delta_z = self.z - self.zw0

        if propagator[1] == "I":
            Cima = 0.0
        else:
            Cima = 1 / delta_z

        self._C = Cima

        if propagator == "II":
            lens_phase = 1.0 / lens_fl
        elif propagator == "IO":
            lens_phase = 1 / lens_fl + Cima
        elif propagator == "OI":
            lens_phase = 1.0 / lens_fl - Cobj
        elif propagator == "OO":
            lens_phase = 1.0 / lens_fl - Cobj + Cima

        x = (np.arange(self._wfo.shape[1]) - self._wfo.shape[1] // 2) * self.dx
        y = (np.arange(self._wfo.shape[0]) - self._wfo.shape[0] // 2) * self.dy

        xx, yy = np.meshgrid(x, y)
        qphase = -(xx**2 + yy**2) * (0.5 * lens_phase / self.wl)

        self._fratio = np.abs(delta_z) / (2 * wz)
        self._wfo = self._wfo * np.exp(2.0j * np.pi * qphase)

    def Magnification(self, My, Mx=None):
        """
        Given the optical magnification along one or both directions, updates the sampling along both directions,
        the beam semi-diameter, the Rayleigh distance, the distance to focus, and the beam focal ratio

        Parameters
        ----------
        My: scalar
            optical magnification along tangential direction
        Mx: scalar
            optical magnification along sagittal direction

        Returns
        -------
        out: None
            updates the wfo parameters

        """
        if Mx is None:
            Mx = My

        assert Mx > 0.0, "Negative magnification not implemented yet."
        assert My > 0.0, "Negative magnification not implemented yet."

        self._dx *= Mx
        self._dy *= My

        if np.abs(Mx - 1.0) < 1.0e-8 or Mx is None:
            logger.trace(
                "Does not do anything if magnification x is close to unity."
            )
            return

        logger.warning(
            "Gaussian beam magnification is implemented, but has not been tested."
        )

        # Current distance to focus (before magnification)
        delta_z = self.z - self.zw0
        # Current w(z) (before magnification)
        wz = self.w0 * np.sqrt(1.0 + ((self.z - self.zw0) / self.zr) ** 2)

        # Apply magnification following ABCD Gaussian beam prescription
        # i.e. w'(z) = Mx*w(z), R'(z) = Mx**2 * R(z)

        delta_z *= Mx**2
        wz *= Mx
        self._w0 *= Mx  # From Eq 56, Lawrence (1992)
        self._zr *= Mx**2
        self._zw0 = self.z - delta_z
        self._fratio = np.abs(delta_z) / (2 * wz)

    def ChangeMedium(self, n1n2):
        """
        Given the ratio of refractive indices n1/n2 for light propagating from a medium with refractive index n1,
        into a medium with refractive index n2, updates the Rayleigh distance, the wavelength, the distance to focus,
        and the beam focal ratio

        Parameters
        ----------
        n1n2

        Returns
        -------
        out: None
            updates the wfo parameters

        """
        _n1n2 = np.abs(n1n2)

        # Current distance to focus (before magnification)
        delta_z = self.z - self.zw0

        delta_z /= n1n2
        self._zr /= n1n2
        self._wl *= n1n2
        self._zw0 = self.z - delta_z
        self._fratio /= n1n2

    def ptp(self, dz):
        """
        Plane-to-plane (far field) wavefront propagator

        Parameters
        ----------
        dz: scalar
            propagation distance
        """
        if np.abs(dz) < 0.001 * self.wl:
            logger.debug(
                "Thickness smaller than 1/1000 wavelength. Returning.."
            )
            return

        if self.C != 0:
            logger.error("PTP wavefront should be planar")
            raise ValueError("PTP wavefront should be planar")

        wf = np.fft.ifftshift(self._wfo)
        wf = np.fft.fft2(wf, norm="ortho")
        fx = np.fft.fftfreq(wf.shape[1], d=self.dx)
        fy = np.fft.fftfreq(wf.shape[0], d=self.dy)
        fxx, fyy = np.meshgrid(fx, fy)
        qphase = (np.pi * self.wl * dz) * (fxx**2 + fyy**2)
        wf = np.fft.ifft2(np.exp(-1.0j * qphase) * wf, norm="ortho")

        self._z = self._z + dz

        self._wfo = np.fft.fftshift(wf)

    def stw(self, dz):
        """
        Spherical-to-waist (near field to far field) wavefront propagator

        Parameters
        ----------
        dz: scalar
            propagation distance
        """
        if np.abs(dz) < 0.001 * self.wl:
            logger.debug(
                "Thickness smaller than 1/1000 wavelength. Returning.."
            )
            return

        if self.C == 0.0:
            logger.error("STW wavefront should not be planar")
            raise ValueError("STW wavefront should not be planar")

        s = "forward" if dz >= 0 else "reverse"

        wf = np.fft.ifftshift(self._wfo)
        if s == "forward":
            wf = np.fft.fft2(wf, norm="ortho")
        elif s == "reverse":
            wf = np.fft.ifft2(wf, norm="ortho")

        fx = np.fft.fftfreq(wf.shape[1], d=self.dx)
        fy = np.fft.fftfreq(wf.shape[0], d=self.dy)
        fxx, fyy = np.meshgrid(fx, fy)

        qphase = (np.pi * self.wl * dz) * (fxx**2 + fyy**2)

        self._z = self._z + dz
        self._C = 0.0
        self._dx = (fx[1] - fx[0]) * self.wl * np.abs(dz)
        self._dy = (fy[1] - fy[0]) * self.wl * np.abs(dz)
        self._wfo = np.fft.fftshift(np.exp(1.0j * qphase) * wf)

    def wts(self, dz):
        """
        Waist-to-spherical (far field to near field) wavefront propagator

        Parameters
        ----------
        dz: scalar
            propagation distance
        """
        if np.abs(dz) < 0.001 * self.wl:
            logger.debug(
                "Thickness smaller than 1/1000 wavelength. Returning.."
            )
            return

        if self.C != 0.0:
            logger.error("WTS wavefront should be planar")
            raise ValueError("WTS wavefront should be planar")

        s = "forward" if dz >= 0 else "reverse"

        x = (np.arange(self._wfo.shape[1]) - self._wfo.shape[1] // 2) * self.dx
        y = (np.arange(self._wfo.shape[0]) - self._wfo.shape[0] // 2) * self.dy

        xx, yy = np.meshgrid(x, y)
        qphase = (np.pi / (dz * self.wl)) * (xx**2 + yy**2)
        wf = np.fft.ifftshift(np.exp(1.0j * qphase) * self._wfo)
        if s == "forward":
            wf = np.fft.fft2(wf, norm="ortho")
        elif s == "reverse":
            wf = np.fft.ifft2(wf, norm="ortho")

        self._z = self._z + dz
        self._C = 1 / (self.z - self.zw0)
        self._dx = self.wl * np.abs(dz) / (wf.shape[1] * self.dx)
        self._dy = self.wl * np.abs(dz) / (wf.shape[1] * self.dy)
        self._wfo = np.fft.fftshift(wf)

    def propagate(self, dz):
        """
        Wavefront propagator. Selects the appropriate propagation primitive and applies the wf propagation

        Parameters
        ----------
        dz: scalar
            propagation distance
        """
        propagator = self.insideout() + self.insideout(self.z + dz)
        z1 = self.z
        z2 = self.z + dz

        if propagator == "II":
            self.ptp(dz)
        elif propagator == "OI":
            self.stw(self.zw0 - z1)
            self.ptp(z2 - self.zw0)
        elif propagator == "IO":
            self.ptp(self.zw0 - z1)
            self.wts(z2 - self.zw0)
        elif propagator == "OO":
            self.stw(self.zw0 - z1)
            self.wts(z2 - self.zw0)

    def zernikes(
        self, index, Z, ordering, normalize, radius, offset=0.0, origin="x"
    ):
        """
        Add a WFE represented by a Zernike expansion

        Parameters
        ----------
        index: array of integers
            Sequence of zernikes to use. It should be a continuous sequence.
        Z : array of floats
            The coefficients of the Zernike polynomials in meters.
        ordering: string
            Can be 'ansi', 'noll', 'fringe', or 'standard'.
        normalize: bool
            Polynomials are normalised to RMS=1 if True, or to unity at radius if False.
        radius: float
            The radius of the circular aperture over which the polynomials are calculated.
        offset: float
            Angular offset in degrees.
        origin: string
            Angles measured counter-clockwise positive from x axis by default (origin='x').
            Set origin='y' for angles measured clockwise-positive from the y-axis.

        Returns
        -------
        out: masked array
            the WFE
        """
        assert not np.any(
            np.diff(index) - 1
        ), "Zernike sequence should be continuous"

        x = (np.arange(self._wfo.shape[1]) - self._wfo.shape[1] // 2) * self.dx
        y = (np.arange(self._wfo.shape[0]) - self._wfo.shape[0] // 2) * self.dy

        xx, yy = np.meshgrid(x, y)
        rho = np.sqrt(xx**2 + yy**2) / radius

        if origin == "x":
            phi = np.arctan2(yy, xx) + np.deg2rad(offset)
        elif origin == "y":
            phi = np.arctan2(xx, yy) + np.deg2rad(offset)
        else:
            logger.error(
                "Origin {} not recognised. Origin shall be either x or y".format(
                    origin
                )
            )
            raise ValueError(
                "Origin {} not recognised. Origin shall be either x or y".format(
                    origin
                )
            )
        zernike = Zernike(
            len(index), rho, phi, ordering=ordering, normalize=normalize
        )
        zer = zernike()
        wfe = (zer.T * Z).T.sum(axis=0)
        self._wfo = self._wfo * np.exp(
            2.0 * np.pi * 1j * wfe / self._wl
        ).filled(0)

        return wfe


if __name__ == "__main__":
    pass



import importlib.metadata as metadata
import os
from datetime import date

from .__version__ import __version__

# load package info
__pkg_name__ = metadata.metadata("paos")["Name"]
__url__ = metadata.metadata("paos")["Home-page"]
__author__ = metadata.metadata("paos")["Author"]
__email__ = metadata.metadata("paos")["Author_email"]
__license__ = metadata.metadata("paos")["license"]
__copyright__ = "2021-{:d}, {}".format(date.today().year, __author__)
__summary__ = metadata.metadata("paos")["Summary"]

try:
    base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
except NameError:
    base_dir = None

__branch__ = None
if base_dir is not None and os.path.exists(os.path.join(base_dir, ".git")):
    git_folder = os.path.join(base_dir, ".git")
    with open(os.path.join(git_folder, "HEAD")) as fp:
        ref = fp.read().strip()
    ref_dir = ref[5:]
    __branch__ = ref[16:]
    try:
        with open(os.path.join(git_folder, ref_dir)) as fp:
            __commit__ = fp.read().strip()
    except FileNotFoundError:
        __commit__ = None
else:
    __commit__ = None

# initialise logger
import logging

logger = logging.getLogger(__pkg_name__)
logger.info(f"code version {__version__}")

from paos.log import setLogLevel

setLogLevel(logging.INFO)

# from paos.log import addLogFile
# addLogFile(level=logging.INFO)

from paos.classes.wfo import WFO
from paos.classes.abcd import ABCD
from paos.classes.zernike import Zernike

from paos.core.parseConfig import parse_config
from paos.core.coordinateBreak import coordinate_break
from paos.core.raytrace import raytrace
from paos.core.plot import plot_pop
from paos.core.saveOutput import save_output, save_datacube
from paos.core.run import run

# initialise plotter
import matplotlib.pyplot as plt

plt.rcParams["figure.facecolor"] = "white"
plt.rc("lines", linewidth=1.5)
plt.rc(
    "axes",
    axisbelow=True,
    titleweight="bold",
    labelcolor="dimgray",
    labelweight="bold",
)
plt.rc("font", size=14)



# This code is inspired by the logger developed for TauREx 3.1.
# Therefore, we attach here the TauREx3.1 license:
#
# BSD 3-Clause License
#
# Copyright (c) 2019, Ahmed F. Al-Refaie, Quentin Changeat, Ingo Waldmann, Giovanna Tinetti
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the names of the copyright holders nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import logging

from .logger import generate_logger_name
from .logger import Logger
from .logger import traced
from paos import __pkg_name__

last_log = logging.INFO

logging.TRACE = 15
logging.addLevelName(logging.TRACE, "TRACE")

def trace(self, message, *args, **kws):
    """
    Log TRACE level.
    Trace level log should be produced anytime a function or a methods is entered and exited."""
    if self.isEnabledFor(logging.TRACE):
        # Yes, logger takes its '*args' as 'args'.
        self._log(logging.TRACE, message, args, **kws)


logging.Logger.trace = trace


def setLogLevel(level, log_id=0):
    """
    Simple function to set the logger level

    Parameters
    ----------
    level: logging level
    log_id: int
        this is the index of the handler to edit. The basic handler index is 0.
        Every added handler is appended to the list. Default is 0.

    """
    global last_log
    from .logger import root_logger

    root_logger.handlers[log_id].setLevel(level)
    last_log = level


def disableLogging(log_id=0):
    """
    It disables the logging setting the log level to ERROR.

    Parameters
    ----------
    log_id: int
        this is the index of the handler to edit. The basic handler index is 0.
        Every added handler is appended to the list. Default is 0.

    """
    setLogLevel(logging.ERROR, log_id)


def enableLogging(level=logging.INFO, log_id=0):
    """
    It disables the logging setting the log level to ERROR.

    Parameters
    ----------
    level: logging level
        Default is logging.INFO.
    log_id: int
        this is the index of the handler to edit. The basic handler index is 0.
        Every added handler is appended to the list. Default is 0.

    """
    global last_log
    if last_log is None:
        last_log = level
    setLogLevel(level, log_id)


def addHandler(handler):
    """
    It adds a handler to the logging handlers list.

    Parameters
    ----------
    handler: logging handler

    """
    from .logger import root_logger

    root_logger.addHandler(handler)


def addLogFile(
    fname="{}.log".format(__pkg_name__), reset=False, level=logging.DEBUG
):
    """
    It adds a log file to the handlers list.

    Parameters
    ----------
    fname: str
        name for the log file. Default is exosim.log.
    reset: bool
        it reset the log file if it exists already. Default is False.
    level: logging level
        Default is logging.INFO.
    """
    if reset:
        import os

        try:
            os.remove(fname)
        except OSError:
            pass
    file_handler = logging.FileHandler(fname)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    file_handler.setFormatter(formatter)
    file_handler.setLevel(level)
    addHandler(file_handler)



import numpy as np

from paos import logger


class ABCD:
    """
    ABCD matrix class for paraxial ray tracing.

    Attributes
    ----------
    thickness: scalar
        optical thickness
    power: scalar
        optical power
    M: scalar
        optical magnification
    n1n2: scalar
        ratio of refractive indices n1/n2 for light propagating
        from a medium with refractive index n1, into a medium
        with refractive index n2
    c : scalar
        speed of light. Can take values +1 for light travelling left-to-right (+Z),
        and -1 for light travelling right-to-left (-Z)

    Note
    ----------
    The class properties can differ from the value of the parameters used at
    class instantiation. This because the ABCD matrix is decomposed into four primitives,
    multiplied together as discussed in :ref:`Optical system equivalent`.

    Examples
    --------

    >>> from paos.classes.abcd import ABCD
    >>> thickness = 2.695  # mm
    >>> radius = 31.850  # mm
    >>> n1, n2 = 1.0, 1.5
    >>> abcd = ABCD(thickness=thickness, curvature=1.0/radius, n1=n1, n2=n2)
    >>> (A, B), (C, D) = abcd.ABCD

    """

    def __init__(self, thickness=0.0, curvature=0.0, n1=1.0, n2=1.0, M=1.0):
        """
        Initialize the ABCD matrix.

        Parameters
        ----------
        thickness: scalar
            optical thickness. It is positive from left to right. Default is 0.0
        curvature: scalar
            inverse of the radius of curvature: it is positive if the center of curvature
            lies on the right. If n1=n2, the parameter is assumed describing
            a thin lens of focal ratio fl=1/curvature. Default is 0.0
        n1: scalar
            refractive index of the first medium. Default is 1.0
        n2: scalar
            refractive index of the second medium. Default is 1.0
        M: scalar
            optical magnification. Default is 1.0

        Note
        -----
        Light is assumed to be propagating from a medium with refractive index n1
        into a medium with refractive index n2.

        Note
        -----
        The refractive indices are assumed to be positive when light propagates
        from left to right (+Z), and negative when light propagates from right
        to left (-Z)
        """

        if n1 == 0 or n2 == 0 or M == 0:
            logger.error(
                "Refractive index and magnification shall not be zero"
            )
            raise ValueError(
                "Refractive index and magnification shall not be zero"
            )

        T = np.array([[1.0, thickness], [0, 1.0]])

        if n1 == n2:
            # Assume a thin lens
            D = np.array([[1.0, 0.0], [-curvature, 1.0]])
        else:
            # Assume dioptre or mirror
            D = np.array([[1.0, 0.0], [-(1 - n1 / n2) * curvature, n1 / n2]])

        M = np.array([[M, 0.0], [0.0, 1.0 / M]])

        self._ABCD = T @ D @ M

        # Remove because not needed and would break ABCD surface type when defined in lens.ini file
        # self._n1 = n1
        # self._n2 = n2
        self._cin = np.sign(n1)
        self._cout = np.sign(n2)

    @property
    def thickness(self):
        (A, B), (C, D) = self._ABCD
        return B / D

    @property
    def M(self):
        (A, B), (C, D) = self._ABCD
        return (A * D - B * C) / D

    @property
    def n1n2(self):
        (A, B), (C, D) = self._ABCD
        return D * self.M

    @property
    def power(self):
        (A, B), (C, D) = self._ABCD
        return -C / self.M

    # @property
    # def n1(self):
    #    return self._n1

    # @property
    # def n2(self):
    #    return self._n2

    @property
    def cin(self):
        return self._cin

    @cin.setter
    def cin(self, c):
        self._cin = c

    @property
    def cout(self):
        return self._cout

    @cout.setter
    def cout(self, c):
        self._cout = c

    @property
    def f_eff(self):
        return 1 / (self.power * self.M)

    @property
    def ABCD(self):
        return self._ABCD

    @ABCD.setter
    def ABCD(self, ABCD):
        self._ABCD = ABCD.copy()

    def __call__(self):
        return self._ABCD

    def __mul__(self, other):
        ABCD_new = self._ABCD @ other()
        out = ABCD()
        out.ABCD = ABCD_new
        out.cin = other.cin
        out.cout = other.cout

        return out



import argparse
import logging
import os

from paos import __version__ as version
from paos import logger
from paos.gui.paosGui import PaosGui
from paos.log import addLogFile
from paos.log import setLogLevel


def main():
    logger.info("code version {}".format(version))

    parser = argparse.ArgumentParser(description="PAOS GUI {}".format(version))
    parser.add_argument(
        "-c",
        "--configuration",
        dest="conf",
        type=str,
        required=False,
        help="Input configuration file to pass",
    )
    parser.add_argument(
        "-d",
        "--debug",
        dest="debug",
        default=False,
        required=False,
        help="enable debug mode",
        action="store_true",
    )
    parser.add_argument(
        "-l",
        "--logger",
        dest="log",
        default=False,
        required=False,
        help="save log file",
        action="store_true",
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="output",
        type=str,
        required=False,
        default=None,
        help="Output file path",
    )

    args = parser.parse_args()
    passvalue = {"conf": args.conf, "debug": args.debug, "output": args.output}

    if args.debug:
        setLogLevel(logging.DEBUG)

    if args.log:
        if isinstance(args.output, str):
            log_filename = "{}/paos.log".format(os.path.dirname(args.output))
            logger.info("log file name: {}".format(log_filename))
            addLogFile(fname=log_filename)
        else:
            addLogFile()

    PaosGui(passvalue=passvalue)()

    logger.info("PAOS GUI exited")
    return


if __name__ == "__main__":
    main()



import datetime
import os
from copy import deepcopy as dc

import h5py
import numpy as np

from paos import __author__
from paos import __version__
from paos import logger


def remove_keys(dictionary, keys):
    """
    Removes item at specified index from dictionary.

    Parameters
    ----------
    dictionary: dict
        input dictionary
    keys:
        keys to remove from the input dictionary

    Returns
    -------
    None
        Updates the input dictionary by removing specific keys

    Examples
    -------
    >>> from paos.core.saveOutput import remove_keys
    >>> my_dict = {'a': 1, 'b': 2, 'c': 3, 'd': 4}
    >>> print(my_dict)
    >>> keys_to_drop = ['a', 'c', 'e']
    >>> remove_keys(my_dict, keys_to_drop)
    >>> print(my_dict)

    """
    for k in keys:
        dictionary.pop(k, "key not found")
    return


def save_recursively_to_hdf5(dictionary, outgroup):
    """
    Given a dictionary and a hdf5 object, saves the dictionary to the hdf5 object.

    Parameters
    ----------
    dictionary: dict
        a dictionary instance to be stored in a hdf5 file
    outgroup
        a hdf5 file object in which to store the dictionary instance

    Returns
    -------
    None
        Save the dictionary recursively to the hdf5 output file

    """

    for key, data in dictionary.items():
        if isinstance(data, dict):
            sub_outgroup = outgroup.create_group(key)
            save_recursively_to_hdf5(data, sub_outgroup)
        elif isinstance(data, (str, int, float, tuple)):
            outgroup.create_dataset(key, data=data)
        elif isinstance(data, np.ndarray):
            outgroup.create_dataset(
                key, data=data, shape=data.shape, dtype=data.dtype
            )
        elif isinstance(data, list):
            asciiList = [n.encode("ascii", "ignore") for n in data]
            outgroup.create_dataset(
                key, data=asciiList, shape=(len(asciiList), 1), dtype="S10"
            )
        elif data is None:
            logger.warning("key {} is None".format(key))
            continue
        else:
            logger.error("data type for {} not supported".format(key))
            raise NameError("data type not supported")


def save_info(file_name, out):
    """
    Inspired by a similar function from ExoRad2.
    Given a hdf5 file name and a hdf5 file object, saves the information about the
    paos package to the hdf5 file object. This information includes the file name,
    the time of creation, the package creator names, the package name, the package
    version, the hdf5 package version and the h5py version.

    Parameters
    ----------
    file_name: str
        the hdf5 file name for saving the POP simulation
    out
        the hdf5 file object

    Returns
    -------
    None
        Saves the paos package information to the hdf5 output file

    """

    attrs = {
        "file_name": file_name,
        "file_time": datetime.datetime.now().isoformat(),
        "creator": __author__,
        "program_name": __package__.upper(),
        "program_version": __version__,
        "HDF5_Version": h5py.version.hdf5_version,
        "h5py_version": h5py.version.version,
    }

    info_group = out.create_group("info")
    save_recursively_to_hdf5(attrs, info_group)

    return


def save_retval(retval, keys_to_keep, out):
    """
    Given the POP simulation output dictionary, the keys to store at each surface and the
    hdf5 file object, it saves the output dictionary to a hdf5 file.

    Parameters
    ----------
    retval: dict
        POP simulation output dictionary to be saved into hdf5 file
    keys_to_keep: list
        dictionary keys to store at each surface. example: ['amplitude', 'dx', 'dy']
    out: `~h5py.File`
        instance of hdf5 file object

    Returns
    -------
    None
        Saves the POP simulation output dictionary to the hdf5 output file

    """

    for index in retval.keys():

        group_name = "S{:02d}".format(index)
        logger.trace("saving {}".format(group_name))

        item = dc(retval[index])

        if item["aperture"] is not None:
            item["aperture"] = item["aperture"].__dict__

        item["ABCDs"] = item["ABCDs"].__dict__
        item["ABCDt"] = item["ABCDt"].__dict__

        if keys_to_keep is None:
            keys_to_keep = item.keys()

        keys_to_drop = [key for key in item.keys() if key not in keys_to_keep]
        remove_keys(item, keys_to_drop)

        outgroup = out.create_group(group_name)
        save_recursively_to_hdf5(item, outgroup)

    return


def save_output(retval, file_name, keys_to_keep=None, overwrite=True):
    """
    Given the POP simulation output dictionary, a hdf5 file name and the keys to store
    at each surface, it saves the output dictionary along with the paos package information
    to the hdf5 output file. If indicated, overwrites past output file.

    Parameters
    ----------
    retval: dict
        POP simulation output dictionary to be saved into hdf5 file
    file_name: str
        the hdf5 file name for saving the POP simulation
    keys_to_keep: list
        dictionary keys to store at each surface. example: ['amplitude', 'dx', 'dy']
    overwrite: bool
        if True, overwrites past output file

    Returns
    -------
    None
        Saves the POP simulation output dictionary along with the paos package information
        to the hdf5 output file

    Examples
    --------

    >>> from paos.core.parseConfig import parse_config
    >>> from paos.core.run import run
    >>> from paos.core.saveOutput import save_output
    >>> pup_diameter, parameters, wavelengths, fields, opt_chains = parse_config('path/to/ini/file')
    >>> ret_val = run(pup_diameter, 1.0e-6 * wavelengths[0], parameters['grid size'],
    >>>              parameters['zoom'], fields[0], opt_chains[0])
    >>> save_output(ret_val, 'path/to/hdf5/file', keys_to_keep=['wfo', 'dx', 'dy'], overwrite=True)

    """
    assert isinstance(retval, dict), "parameter retval must be a dict"
    assert isinstance(file_name, str), "parameter file_name must be a string"
    if keys_to_keep is not None:
        assert isinstance(
            keys_to_keep, list
        ), "parameter keys_to_keep must be a list of strings"

    logger.info("saving {} started...".format(file_name))

    if overwrite:
        logger.info("removing old file")
        if os.path.isfile(file_name):
            os.remove(file_name)

    with h5py.File(file_name, "a") as out:

        save_info(file_name, out)
        save_retval(retval, keys_to_keep, out)

    logger.info("saving ended.")

    return


def save_datacube(
    retval_list, file_name, group_names, keys_to_keep=None, overwrite=True
):
    """
    Given a list of dictionaries with POP simulation output, a hdf5 file name, a list of
    identifiers to tag each simulation and the keys to store at each surface, it saves the
    outputs to a data cube along with the paos package information to the hdf5 output file.
    If indicated, overwrites past output file.

    Parameters
    ----------
    retval_list: list
        list of dictionaries with POP simulation outputs to be saved into a single hdf5 file
    file_name: str
        the hdf5 file name for saving the POP simulation
    group_names: list
        list of strings with unique identifiers for each POP simulation. example: for one
        optical chain run at different wavelengths, use each wavelength as identifier.
    keys_to_keep: list
        dictionary keys to store at each surface. example: ['amplitude', 'dx', 'dy]
    overwrite: bool
        if True, overwrites past output file

    Returns
    -------
    None
        Saves a list of dictionaries with the POP simulation outputs to a single hdf5 file
        as a datacube with group tags (e.g. the wavelengths) to identify each simulation,
        along with the paos package information.

    Examples
    --------

    >>> from paos.core.parseConfig import parse_config
    >>> from paos.core.run import run
    >>> from paos.core.saveOutput import save_datacube
    >>> from joblib import Parallel, delayed
    >>> from tqdm import tqdm
    >>> pup_diameter, parameters, wavelengths, fields, opt_chains = parse_config('path/to/ini/file')
    >>> ret_val_list = Parallel(n_jobs=2)(delayed(run)(pup_diameter, 1.0e-6 * wl, parameters['grid size'],
    >>>               parameters['zoom'], fields[0], opt_chains[0]) for wl in tqdm(wavelengths))
    >>> group_tags = list(map(str, wavelengths))
    >>> save_datacube(ret_val_list, 'path/to/hdf5/file', group_tags,
    >>>               keys_to_keep=['amplitude', 'dx', 'dy'], overwrite=True)

    """

    assert isinstance(
        retval_list, list
    ), "parameter retval_list must be a list"
    assert isinstance(file_name, str), "parameter file_name must be a string"
    assert isinstance(
        group_names, list
    ), "parameter group_names must be a list of strings"

    if keys_to_keep is not None:
        assert isinstance(
            keys_to_keep, list
        ), "parameter keys_to_keep must be a list of strings"

    logger.info("Saving {} started...".format(file_name))

    if overwrite:
        logger.info("Remove old file")
        if os.path.exists(file_name) and os.path.isfile(file_name):
            os.remove(file_name)

    with h5py.File(file_name, "a") as cube:

        save_info(file_name, cube)

        for group_name, retval in zip(group_names, retval_list):

            out = cube.create_group(group_name)
            logger.trace("saving group {}".format(out))

            save_retval(retval, keys_to_keep, out)

    logger.info("Saving ended.")

    return



import matplotlib.pyplot as plt
import numpy as np

from paos import logger
from paos.core.plot import do_legend


class Material:
    """
    Class for handling different optical materials for use in `PAOS`
    """

    def __init__(self, wl, Tambient=-218.0, Pambient=1.0, materials=None):
        """
        Parameters
        ----------
        Tambient: scalar
            Ambient temperature during operation (:math:`^{\\circ} C`)
        Pambient: scalar
            Ambient pressure during operation (atm)
        wl: scalar or array
            wavelength in microns
        materials: dict
            library of materials for optical use
        """
        self.wl = wl
        self.Tambient = Tambient
        self.Pambient = Pambient

        if materials is None:
            from .lib import materials

            logger.debug("Using default library of optical materials")
        self.materials = materials

    def sellmeier(self, par):
        """
        Implements the Sellmeier 1 equation to estimate the glass index of refraction relative to air at the glass
        reference temperature :math:`T_{ref} = 20 ^{\\circ}C` and pressure :math:`P_{ref} = 1 \\ atm`.

        The Sellmeier 1 equation consists of three terms and is given as
        :math:`\\displaystyle n^{2}(\\lambda )=1+{\\frac {K_{1}\\lambda ^{2}}{\\lambda ^{2}-L_{1}}}+
        {\\frac {K_{2}\\lambda ^{2}}{\\lambda ^{2}-L_{2}}}+{\\frac {K_{3}\\lambda ^{2}}{\\lambda ^{2}-L_{3}}}`

        Parameters
        ----------
        par: dict
            dictionary containing the :math:`K_1`, :math:`L_1`, :math:`K_2`, :math:`L_2`, :math:`K_3`, :math:`L_3`
            parameters of the Sellmeier 1 model

        Returns
        -------
        out: scalar or array (same shape as wl)
            the refractive index
        """
        wl2 = self.wl**2
        n2_1 = par["K1"] * wl2 / (wl2 - par["L1"])
        n2_1 += par["K2"] * wl2 / (wl2 - par["L2"])
        n2_1 += par["K3"] * wl2 / (wl2 - par["L3"])

        return np.sqrt(n2_1 + 1.0)

    @staticmethod
    def nT(n, D0, delta_T):
        """
        Estimate the change in the glass absolute index of refraction with temperature as

        :math:`n(\\Delta T) = \\Delta n_{abs} + n`

        where

        :math:`\\Delta n_{abs} = \\frac{n^2 - 1}{2 n} D_0 \\Delta T`

        Parameters
        ----------
        n: scalar or array
            relative index at the reference temperature of the glass
        D0: scalar
            model parameter (constant provided by the glass manufacturer to describe the glass thermal behaviour)
        delta_T: scalar
            change in temperature relative to the reference temperature of the glass.
            It is positive if the temperature is greater than the reference temperature of the glass

        Returns
        -------
        out: scalar or array (same shape as n)
            the scaled relative index
        """
        dnabs = (n**2 - 1.0) / (2.0 * n) * D0 * delta_T

        return n + dnabs

    def nair(self, T, P=1.0):
        """
        Estimate the air index of refraction at wavelength :math:`\\lambda`, temperature :math:`T`,
        and relative pressure :math:`P` as

        :math:`n_{air} = 1 + \\frac{\\left(n_{ref} - 1\\right) P}{1.0 + (T - 15) \\cdot (3.4785 \\times 10^{-3}) }`

        where

        :math:`n_{ref} = 1 + \\left[6432.8 + \\frac{2949810 \\lambda^{2}}{146 \\lambda^{2} - 1} + \\frac{25540
        \\lambda^{2}}{41 \\lambda^{2} - 1} \\right] \\cdot 10^{-8}`

        This formula for the index of air is from F. Kohlrausch, Praktische Physik, 1968, Vol 1, page 408.

        Parameters
        ----------
        T: scalar
            temperature in :math:`^{\\circ} C`
        P: scalar
            relative pressure in atmospheres (dimensionless in the formula). Defaults to 1 atm.

        Returns
        -------
        out: scalar or array (same shape as wl)
            the air index of refraction

        Note
        ----
        1) Air at the system temperature and pressure is defined to be 1.0, all other indices are relative
        2) ``PAOS`` can easily model systems used in a vacuum by changing the air pressure to zero
        """

        wl2 = self.wl**2

        nref = 1.0 + 1.0e-8 * (
            6432.8
            + 2949810.0 * wl2 / (146.0 * wl2 - 1.0)
            + 25540.0 * wl2 / (41.0 * wl2 - 1.0)
        )
        nair = 1.0 + (nref - 1.0) * P / (1.0 + 3.4785e-3 * (T - 15))

        return nair

    def nmat(self, name):
        """
        Given the name of an optical glass, returns the index of refraction in vacuum as a function of wavelength.

        Parameters
        ----------
        name: str
            name of the optical glass

        Returns
        -------
        out: tuple(scalar, scalar) or tuple(array, array)
            returns two arrays for the glass index of refraction at the given wavelengths:
            the index of refraction at :math:`T_{ref} = 20^{\\circ} C` (nmat0)
            and the index of refraction at :math:`T_{amb}` (nmat)
        """

        name = name.upper()
        if name not in self.materials.keys():
            logger.error(f"Glass {name} currently not supported.")

        material = self.materials[name]
        logger.debug(f'Glass name: {name} -- T ref: {material["Tref"]}')

        nmat0 = self.sellmeier(par=material["sellmeier"]) * self.nair(
            T=material["Tref"], P=self.Pambient
        )
        nmat = self.nT(
            n=nmat0,
            D0=material["Tmodel"]["D0"],
            delta_T=self.Tambient - material["Tref"],
        )

        return nmat0, nmat

    def plot_relative_index(self, material_list=None, ncols=2, figname=None):
        """
        Given a list of materials for optical use, plots the relative index in function of wavelength,
        at the reference and operating temperature.

        Parameters
        ----------
        material_list: list
            a list of materials, e.g. ['SF11', 'ZNSE']
        ncols: int
            number of columns for the subplots
        figname: str
            name of figure to save

        Returns
        -------
        out: None
            displays the plot output or stores it to the indicated plot path

        Examples
        --------

        >>> from paos.util.material import Material
        >>> Material(wl = np.linspace(1.8, 8.0, 1024)).plot_relative_index(material_list=['Caf2', 'Sf11', 'Sapphire'])

        """
        i, j = None, None

        if material_list is None:
            material_list = []

        n_subplots = len(material_list)
        if ncols > n_subplots:
            ncols = n_subplots

        nrows = n_subplots // ncols
        if n_subplots % ncols:
            nrows += 1

        figsize = (8 * ncols, 6 * nrows)
        fig, ax = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize)
        fig.patch.set_facecolor("white")
        plt.subplots_adjust(hspace=0.3, wspace=0.5)

        for k, name in enumerate(material_list):
            if n_subplots == 1:
                axis = ax
            elif n_subplots == 2:
                axis = ax[k]
            else:
                i = k % ncols
                j = k // ncols
                axis = ax[j, i]

            nmat0, nmat = self.nmat(name)

            axis.plot(self.wl, nmat0, "--", label=r"T$_{ref}$")
            axis.plot(self.wl, nmat, label=r"T$_{oper}$")
            axis.set_title(name)
            do_legend(axis=axis, ncol=2)
            axis.set_xlabel("Wavelength [micron]")
            axis.set_ylabel("Relative index")
            axis.grid()

            if n_subplots % ncols and k == n_subplots - 1:
                for col in range(i + 1, ncols):
                    ax[j, col].set_visible(False)

        if figname is not None:
            fig.savefig(figname, bbox_inches="tight", dpi=150)
            plt.close()
        else:
            fig.tight_layout()
            plt.show()

        return



import gc
from copy import deepcopy

import numpy as np

from paos import logger
from paos.classes.abcd import ABCD
from paos.classes.wfo import WFO
from paos.core.coordinateBreak import coordinate_break


def push_results(wfo):
    retval = {
        "amplitude": wfo.amplitude,
        "wz": wfo.wz,
        "distancetofocus": wfo.distancetofocus,
        "fratio": wfo.fratio,
        "phase": wfo.phase,
        "dx": wfo.dx,
        "dy": wfo.dy,
        "wfo": wfo.wfo,
        "wl": wfo.wl,
        "extent": wfo.extent,
    }

    return retval


def run(pupil_diameter, wavelength, gridsize, zoom, field, opt_chain):
    """
    Run the POP.

    Parameters
    ----------
    pupil_diameter: scalar
        input pupil diameter in meters
    wavelength: scalar
        wavelength in meters
    gridsize: scalar
        the size of the simulation grid. It has to be a power of 2
    zoom: scalar
        zoom factor
    field: dictionary
        contains the slopes in the tangential and sagittal planes as field={'vt': slopey, 'vs': slopex}
    opt_chain: list
        the list of the optical elements parsed by paos.core.parseConfig.parse_config

    Returns
    -------
    out: dict
        dictionary containing the results of the POP

    Examples
    --------

    >>> from paos.core.parseConfig import parse_config
    >>> from paos.core.run import run
    >>> from paos.core.plot import simple_plot
    >>> pup_diameter, parameters, wavelengths, fields, opt_chains = parse_config('path/to/conf/file')
    >>> ret_val = run(pup_diameter, 1.0e-6 * wavelengths[0], parameters['grid_size'], parameters['zoom'], fields[0], opt_chains[0])

    """

    assert isinstance(opt_chain, dict), "opt_chain must be a dict"

    retval = {}

    vt = np.array([0.0, field["ut"]])
    vs = np.array([0.0, field["us"]])

    ABCDt = ABCD()
    ABCDs = ABCD()

    wfo = WFO(pupil_diameter, wavelength, gridsize, zoom)

    for index, item in opt_chain.items():

        logger.trace("Surface: {}".format(item["name"]))

        if item["type"] == "Coordinate Break":
            logger.trace("Apply coordinate break.")
            vt, vs = coordinate_break(
                vt,
                vs,
                item["xdec"],
                item["ydec"],
                item["xrot"],
                item["yrot"],
                0.0,
            )

        _retval_ = {"aperture": None}

        # Check if aperture needs to be applied
        if "aperture" in item:
            xdec = (
                item["aperture"]["xc"]
                if np.isfinite(item["aperture"]["xc"])
                else vs[0]
            )
            ydec = (
                item["aperture"]["yc"]
                if np.isfinite(item["aperture"]["yc"])
                else vt[0]
            )
            xrad = item["aperture"]["xrad"]
            yrad = item["aperture"]["yrad"]
            xrad *= np.sqrt(1 / (vs[1] ** 2 + 1))
            yrad *= np.sqrt(1 / (vt[1] ** 2 + 1))
            xaper = xdec - vs[0]
            yaper = ydec - vt[0]

            obscuration = (
                False if item["aperture"]["type"] == "aperture" else True
            )

            if np.all(np.isfinite([xrad, yrad])):
                logger.trace("Apply aperture")
                aper = wfo.aperture(
                    xaper,
                    yaper,
                    hx=xrad,
                    hy=yrad,
                    shape=item["aperture"]["shape"],
                    obscuration=obscuration,
                )
                _retval_["aperture"] = aper

        # if item['type'] in ['Standard', 'Paraxial Lens', 'Slit', 'Obscuration']:
        #     xdec = item['xdec'] if np.isfinite(item['xdec']) else vs[0]
        #     ydec = item['ydec'] if np.isfinite(item['ydec']) else vt[0]
        #     xrad = item['xrad']
        #     yrad = item['yrad']
        #     xrad *= np.sqrt(1 / (vs[1] ** 2 + 1))
        #     yrad *= np.sqrt(1 / (vt[1] ** 2 + 1))
        #     xaper = xdec - vs[0]
        #     yaper = ydec - vt[0]
        #
        #     aperture_shape = 'rectangular' if item['type'] == 'Slit' else 'elliptical'
        #     obscuration = True if item['type'] == 'Obscuration' else False
        #     if np.all(np.isfinite([xrad, yrad])):
        #         logger.trace('Apply aperture')
        #         aper = wfo.aperture(xaper, yaper, hx=xrad, hy=yrad,
        #                             shape=aperture_shape, obscuration=obscuration)
        #         _retval_['aperture'] = aper

        # Check if this is a stop surface
        if item["is_stop"]:
            logger.trace("Apply stop")
            wfo.make_stop()

        if item["type"] == "Zernike":
            radius = (
                item["Zradius"] if np.isfinite(item["Zradius"]) else wfo.wz
            )
            wfo.zernikes(
                item["Zindex"],
                item["Z"],
                item["Zordering"],
                item["Znormalize"],
                radius,
                origin=item["Zorigin"],
            )

        _retval_.update(push_results(wfo))

        Ms = item["ABCDs"].M
        Mt = item["ABCDt"].M

        fl = (
            np.inf
            if (item["ABCDt"].power == 0)
            else item["ABCDt"].cout / item["ABCDt"].power
        )
        T = item["ABCDt"].cout * item["ABCDt"].thickness
        n1n2 = item["ABCDt"].n1n2
        logger.trace("n1n2: {:.4f}".format(n1n2))

        if Mt != 1.0 or Ms != 1.0:
            logger.trace("Apply magnification")
            wfo.Magnification(Mt, Ms)

        if np.abs(n1n2) != 1.0:
            logger.trace("Apply medium change")
            wfo.ChangeMedium(n1n2)

        if np.isfinite(fl):
            logger.trace("Apply lens")
            wfo.lens(fl)

        if np.isfinite(T) and np.abs(T) > 1e-10:
            logger.trace("Apply propagation thickness: T: {:.4f}".format(T))
            wfo.propagate(T)

        vt = item["ABCDt"]() @ vt
        vs = item["ABCDs"]() @ vs
        ABCDt = item["ABCDt"] * ABCDt
        ABCDs = item["ABCDs"] * ABCDs

        logger.debug(
            "F num: {:2f}, distance to focus: {:.6f}".format(
                _retval_["fratio"], wfo.distancetofocus
            )
        )

        _retval_["ABCDt"] = ABCDt
        _retval_["ABCDs"] = ABCDs

        if item["save"]:
            logger.trace("Save optical surface to output dict")
            retval[item["num"]] = deepcopy(_retval_)
        del _retval_

    _ = gc.collect()

    return retval



import numpy as np
from scipy.spatial.transform import Rotation as R

from paos import logger


def coordinate_break(vt, vs, xdec, ydec, xrot, yrot, zrot, order=0):
    """
    Performs a coordinate break and estimates the new :math:`\\vec{v_{t}}=(y, u_{y})`
    and :math:`\\vec{v_{s}}=(x, u_{x})`.

    Parameters
    ----------
    vt: array
        vector :math:`\\vec{v_{t}}=(y, u_{y})` describing a ray propagating in the tangential plane
    vs: array
        vector :math:`\\vec{v_{s}}=(x, u_{x})` describing a ray propagating in the sagittal plane
    xdec: float
        x coordinate of the decenter to be applied
    ydec: float
        y coordinate of the decenter to be applied
    xrot: float
        tilt angle around the X axis to be applied
    yrot: float
        tilt angle around the Y axis to be applied
    zrot: float
        tilt angle around the Z axis to be applied
    order: int
        order of the coordinate break, defaults to 0.

    Returns
    -------
    tuple
        two arrays representing the new :math:`\\vec{v_{t}}=(y, u_{y})`
        and :math:`\\vec{v_{s}}=(x, u_{x})`.

    Note
    ----
    When order=0, first a coordinate decenter is applied, followed by a XYZ rotation.
    Coordinate break orders other than 0 not implemented yet.

    """

    if order != 0:
        logger.error(
            "Coordinate break orders other than 0 not implemented yet"
        )
        raise ValueError(
            "Coordinate break orders other than 0 not implemented yet"
        )

    if not np.isfinite(xdec):
        xdec = 0.0
    if not np.isfinite(ydec):
        ydec = 0.0
    if not np.isfinite(xrot):
        xrot = 0.0
    if not np.isfinite(yrot):
        yrot = 0.0
    if not np.isfinite(zrot):
        zrot = 0.0

    # Rotation matrix, intrinsic
    U = R.from_euler("xyz", [xrot, yrot, zrot], degrees=True)

    r0 = [vs[0] - xdec, vt[0] - ydec, 0.0]
    n0 = [vs[1], vt[1], 1]
    n1 = U.inv().apply(n0)
    n1 /= n1[2]
    r1_ln1 = U.inv().apply(r0)
    r1 = r1_ln1 - n1 * r1_ln1[2] / n1[2]

    vt1 = np.array([r1[1], n1[1]])
    vs1 = np.array([r1[0], n1[0]])

    return vt1, vs1



import itertools
import re
import sys

from PySimpleGUI import Button
from PySimpleGUI import Column
from PySimpleGUI import Frame
from PySimpleGUI import popup
from PySimpleGUI import RELIEF_SUNKEN
from PySimpleGUI import Submit
from PySimpleGUI import Text
from PySimpleGUI import TIMEOUT_KEY
from PySimpleGUI import Window
from PySimpleGUI import WINDOW_CLOSE_ATTEMPTED_EVENT
from PySimpleGUI import WINDOW_CLOSED

from paos import logger
from paos import Zernike
from paos.gui.simpleGui import SimpleGui


class ZernikeGui(SimpleGui):
    """
    Generates the Zernike editor for the main ``PAOS`` GUI as a secondary GUI window

    Parameters
    ----------
    config: :class:`~configparser.ConfigParser` object
        the main GUI window parsed configuration file
    values: dict
        the main GUI window dict of values (returned when the Window element is read)
    row: int
        the Zernike surface row index in the lens data editor
    key: str
        the Zernike surface key in the configuration file

    """

    def __init__(self, config, values, row, key):
        super().__init__()

        self.config = config
        self.values = values
        self.row = row
        self.key = key

        # ------ Instantiate global variables ------ #
        self.window = None
        self.zernike = {}
        self.ordering = None
        self.par = ["", "", "", "", ""]
        self.names = ["Par1", "Par2", "Par3", "Par4", "Par5"]
        self.headings = ["Zindex", "Z", "m", "n"]
        self.disabled_cols = [True, False, True, True]
        self.max_rows = None

    def add_row(self, row, dictionary, ordering):
        """

        Parameters
        ----------
        row: int
            the last row index in the Zernike tab
        dictionary: dict
            the dictionary with the Zindex and Z coefficients to update with the new Zernike table row
        ordering: str
            the Zernike coefficients ordering

        Returns
        -------
        out: tuple(int, tuple(int, int))
            Adds a row in the Zernike tab and returns the updated number of rows and the azimuthal number, m,
            and radial number, n for the Zernike coefficients

        """
        row += 1

        # Update the azimuthal and radial number for the Zernike coefficients
        m, n = Zernike.j2mn(N=row, ordering=ordering)
        dictionary["zindex"].append(str(row))
        dictionary["z"].append("0.0")

        # Define the input list to fill the new table row with
        input_list = [str(row - 1), "0.0", m[row - 1], n[row - 1]]

        # Extend the Column layout
        self.window.extend_layout(
            self.window["zernike"],
            [
                self.chain_widgets(
                    row=row,
                    input_list=input_list,
                    prefix="z",
                    disabled_list=self.disabled_cols,
                )
            ],
        )
        # Update the GUI Zernike tab
        self.window["zernike"].update()

        return row, (m, n)

    def make_window(self):
        """
        Generates the Zernike GUI window (secondary window to the main ``PAOS`` GUI window).
        The layout is composed of only one Tab, with four headers: 'Zindex', 'Z', 'm', 'n'.
        'Zindex' contains the index j for each Zernike coefficient (starting from 0), 'Z' contains the Zernike
        coefficients, 'm' contains the azimuthal number and 'n' the radial number for the Zernike coefficients

        Returns
        -------
        out: None
            Generates the Zernike GUI window
        """

        # Get the Zindex and Z coefficients from the parsed configuration file
        if self.key in self.config.keys() and {"zindex", "z"}.issubset(
            self.config[self.key].keys()
        ):
            zindex, z = (
                self.config[self.key]["zindex"],
                self.config[self.key]["z"],
            )
            self.zernike["zindex"] = zindex.split(",") if zindex != "" else ["0"]
            self.zernike["z"] = z.split(",") if z != "" else ["0"]
        else:
            self.zernike["zindex"] = ["0"]
            self.zernike["z"] = ["0"]
        self.max_rows = len(self.zernike["zindex"])
        if len(self.zernike["z"]) != self.max_rows:
            logger.critical(
                "Input zernike index and zernike coefficients differ in length. Quitting..."
            )
            sys.exit()

        # Get the Zernike parameters from the main GUI window (wavelength, ordering, normalization, radius of support
        # aperture of the polynomials and origin, which can be x (counterclockwise positive from x-axis) or y
        # (clockwise positive from y-axis)
        for c, head in enumerate(self.names):
            par_key = "{}_({},{})".format(head, self.row, c + 9)
            if par_key in self.values.keys() and self.values[par_key] != "":
                self.par[c] = self.values[par_key]
        wavelength, self.ordering, normalization, radius, origin = self.par
        if self.ordering == "":
            logger.debug(
                "Zernike ordering is not defined. Defaulting to ordering=standard."
            )
            self.ordering = "standard"
        assert self.ordering in [
            "ansi",
            "standard",
            "noll",
            "fringe",
        ], f"ordering {self.ordering} not supported"

        # ------ Instantiate the azimuthal and radial Zernike order ------ #
        m, n = Zernike.j2mn(N=self.max_rows, ordering=self.ordering)

        # ------ Define the Zernike tab layout ------ #
        layout = [
            [
                Frame(
                    "Parameters",
                    layout=[
                        list(
                            itertools.chain(
                                [
                                    Text(
                                        "Wavelength: {}".format(wavelength),
                                        key="wavelength",
                                    )
                                ],
                                [Text("", size=(6, 1))],
                                [
                                    Text(
                                        "Ordering: {}".format(self.ordering),
                                        key="ordering",
                                    )
                                ],
                                [Text("", size=(6, 1))],
                                [
                                    Text(
                                        "Normalization: {}".format(normalization),
                                        key="normalization",
                                    )
                                ],
                                [Text("", size=(6, 1))],
                                [
                                    Text(
                                        "Radius of S.A.: {}".format(radius),
                                        key="radius",
                                    )
                                ],
                                [Text("", size=(6, 1))],
                                [
                                    Text(
                                        "Origin: {}".format(origin),
                                        key="origin",
                                    )
                                ],
                            )
                        )
                    ],
                    font=self.font_titles,
                    relief=RELIEF_SUNKEN,
                    key="-ZERNIKE MEMO FRAME-",
                )
            ],
            [Text("", size=(10, 2))],
            [
                Frame(
                    "Zernike Setup",
                    layout=[
                        [Text("", size=(24, 1))],
                        [
                            Column(
                                layout=list(
                                    itertools.chain(
                                        [self.add_heading(self.headings)],
                                        [
                                            self.chain_widgets(
                                                row=i + 1,
                                                input_list=[
                                                    r,
                                                    float(self.zernike["z"][i]),
                                                    int(m[i]),
                                                    int(n[i]),
                                                ],
                                                prefix="z",
                                                disabled_list=self.disabled_cols,
                                            )
                                            for i, r in enumerate(
                                                self.zernike["zindex"]
                                            )
                                        ],
                                    )
                                ),
                                scrollable=True,
                                vertical_scroll_only=True,
                                expand_y=True,
                                key="zernike",
                            )
                        ],
                        [Text("", size=(10, 2))],
                        [
                            Frame(
                                "Zernike Actions",
                                layout=[
                                    [
                                        Text("Add a new row: ", size=(24, 1)),
                                        Button(
                                            tooltip="Click to add a new row",
                                            button_text="Add row",
                                            enable_events=True,
                                            key="-ADD ZERNIKE ROW-",
                                        ),
                                    ],
                                    [
                                        Text(
                                            "Add or complete a order: ",
                                            size=(24, 1),
                                        ),
                                        Button(
                                            tooltip="Click to add a radial order",
                                            button_text="Add/Complete order",
                                            enable_events=True,
                                            key="-ADD ZERNIKE RADIAL ORDER-",
                                        ),
                                    ],
                                    [
                                        Text(
                                            "Paste Zernike coefficients: ",
                                            size=(24, 1),
                                        ),
                                        Button(
                                            tooltip="Click to paste Zernike coefficients",
                                            button_text="Paste coefficients",
                                            enable_events=True,
                                            key="PASTE ZERNIKES",
                                        ),
                                    ],
                                ],
                                font=self.font_titles,
                                relief=RELIEF_SUNKEN,
                                key="-ZERNIKE ACTIONS FRAME-",
                            )
                        ],
                    ],
                    font=self.font_titles,
                    relief=RELIEF_SUNKEN,
                    key="ZERNIKE FRAME",
                    expand_y=True,
                )
            ],
            [
                # Submit(
                #     tooltip="Click to submit (debug)", key="-SUBMIT ZERNIKES-"
                # ),
                Button("Exit", key="-EXIT ZERNIKES-"),
            ],
        ]

        # ------ Window creation ------ #
        self.window = Window(
            "Zernike window",
            layout,
            default_element_size=(12, 1),
            return_keyboard_events=True,
            finalize=True,
            enable_close_attempted_event=True,
            resizable=True,
            element_justification="center",
            keep_on_top=False,
        )

    def __call__(self):
        """
        Returns a rendering of the GUI Zernike window and handles the event loop.
        """

        self.make_window()

        # ------ Instantiate local variables ------ #
        order_closed = False

        while True:
            event, values = self.window.read(timeout=1000)
            if event == TIMEOUT_KEY:
                continue
            logger.trace("============ Event = {} ==============".format(event))
            elem = self.window.find_element_with_focus()
            elem_key = (
                elem.Key
                if (elem is not None and isinstance(elem.Key, (str, tuple)))
                else (0, 0)
            )

            # ------- Move with arrow keys within the Zernike tab ------#
            if isinstance(elem_key, str):
                _ = self.move_with_arrow_keys(
                    self.window,
                    event,
                    values,
                    elem_key,
                    self.max_rows,
                    len(self.headings),
                )

            # ------- Properly close the Zernike window ------#
            if event == "-EXIT ZERNIKES-" or event in (
                WINDOW_CLOSE_ATTEMPTED_EVENT,
                WINDOW_CLOSED,
            ):
                self.close_window()
                break

            # ------- Add a new row in the Zernike tab below those already present ------#
            elif event == "-ADD ZERNIKE ROW-":
                # Add row
                self.max_rows, _ = self.add_row(
                    row=self.max_rows,
                    dictionary=self.zernike,
                    ordering=self.ordering,
                )
                # Update the 'zernike' Column scrollbar
                self.update_column_scrollbar(window=self.window, col_key="zernike")

            # ------- Paste from the clipboard to the desired Zernike coefficients 'Z' input cell and below ------#
            elif event == "PASTE ZERNIKES":
                # Find current position in the Zernike tab
                row, col = re.findall("[0-9]+", elem_key.partition("_")[-1])
                if self.headings[int(col)] != "Z":
                    logger.error(
                        "The user shall select any cell from from the Z column. Skipping.."
                    )
                    continue
                # Get the text from the clipboard
                text = self.get_clipboard_text()
                # Update the Z coefficients by pasting and adding new rows
                row0 = row = int(row)
                for text_item in text:
                    self.window["z_({},1)".format(row)].update(text_item)
                    if self.max_rows <= row < row0 + len(text) - 1:
                        self.max_rows, _ = self.add_row(
                            row=self.max_rows,
                            dictionary=self.zernike,
                            ordering=self.ordering,
                        )
                    row += 1
                # Update the Zernike tab scrollbar
                self.update_column_scrollbar(window=self.window, col_key="zernike")

            # ------- Add a radial order to the current Zernike tab by adding rows until the order is closed ------#
            elif event == "-ADD ZERNIKE RADIAL ORDER-":
                # Check ordering
                if self.ordering not in ["ansi", "standard"]:
                    logger.error(
                        "Not supported with {} as ordering. Skipping..".format(
                            self.ordering.capitalize()
                        )
                    )
                    continue
                # Get azimuthal and radial orders and closing order condition
                m, n = Zernike.j2mn(N=self.max_rows, ordering=self.ordering)
                if self.ordering == "standard":
                    order_closed = min(m) == -max(n)
                elif self.ordering == "ansi":
                    order_closed = max(m) == max(n)
                # Keep adding rows to the current Zernike tab until the current radial order (unclosed) is complete
                if not order_closed:
                    while not order_closed:
                        self.max_rows, (m, n) = self.add_row(
                            row=self.max_rows,
                            dictionary=self.zernike,
                            ordering=self.ordering,
                        )
                        if self.ordering == "standard":
                            order_closed = min(m) == -max(n)
                        elif self.ordering == "ansi":
                            order_closed = max(m) == max(n)
                    # When finished, exit
                    continue
                # If the current radial order is closed, keep adding rows to add a new complete radial order
                new_m = new_n = max(n) + 1
                if self.ordering == "standard":
                    new_m = -new_m
                jmax = Zernike.mn2j(m=new_m, n=new_n, ordering=self.ordering)
                while self.max_rows < jmax + 1:
                    self.max_rows, _ = self.add_row(
                        row=self.max_rows,
                        dictionary=self.zernike,
                        ordering=self.ordering,
                    )
                # Update the Zernike tab scrollbar
                self.update_column_scrollbar(window=self.window, col_key="zernike")

            # ------- Display a popup window with the Zernike GUI values given as a flat dictionary ------#
            elif event == "-SUBMIT ZERNIKES-":
                popup(
                    "Paos GUI Zernike window",
                    'You clicked on the "{}" button'.format(event),
                    "The values are",
                    values,
                    keep_on_top=True,
                )

        # ------- Return the updated Zernike dictionary based on the current Zernike tab contents ------#
        zernike = {"zindex": [], "z": []}
        for key, item in values.items():
            row, col = re.findall("[0-9]+", key.partition("_")[-1])
            if col == "0":
                # Column 'Zindex'
                zernike["zindex"].append(item)
            elif col == "1":
                # Column 'Z'
                zernike["z"].append(item)
        zernike["zindex"] = ",".join(zernike["zindex"])
        zernike["z"] = ",".join(zernike["z"])
        zernike["ordering"] = self.ordering

        return zernike



import numpy as np
from matplotlib import pyplot as plt
from matplotlib import ticker as ticks
from matplotlib.patches import Circle
from matplotlib.patches import Ellipse
from matplotlib.patches import Rectangle
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.special import j1

from paos import logger


def do_legend(axis, ncol=1):
    """
    Create a nice legend for the plots

    Parameters
    ----------
    axis: :class:`~matplotlib.axes.Axes`
        An instance of matplotlib axis
    ncol: int
        The number of legend columns

    Returns
    -------
    out: None
        Produces a nice matplotlib legend

    """
    legend = axis.legend(
        loc="best", ncol=ncol, frameon=True, prop={"size": 12}
    )
    legend.get_frame().set_facecolor("white")
    legend.get_frame().set_edgecolor("white")
    legend.get_frame().set_alpha(0.8)
    return


def simple_plot(
    fig,
    axis,
    key,
    item,
    ima_scale,
    origin="lower",
    cmap="viridis",
    options={},
):
    """
    Given the POP simulation output dict, plots the squared amplitude of the
    wavefront at the given optical surface.

    Parameters
    ----------
    fig: :class:`~matplotlib.figure.Figure`
        instance of matplotlib figure artist
    axis: :class:`~matplotlib.axes.Axes`
        instance of matplotlib axes artist
    key: int
        optical surface index
    item: dict
        optical surface dict
    ima_scale: str
        plot color map scale, can be either 'linear' or 'log'
    origin: str
        matplotlib plot origin. Defaults to 'lower'
    cmap: str
        matplotlib plot color map. Defaults to 'viridis'
    options: dict
        dictionary containing the options to override the plotting default for one or more surfaces, specified by the
        dictionary key. Available options are the surface scale, an option to display physical units, the surface
        zoom(out), the plot scale and whether to plot dark rings in correspondance to the zeros of the Airy diffraction
        pattern.
        Examples:
        0) options={4: {'ima_scale':'linear'}}
        1) options={4: {'surface_scale':60, 'ima_scale':'linear'}}
        2) options={4: {'surface_scale':21, 'pixel_units':True, 'ima_scale':'linear'}}
        3) options={4: {'surface_zoom':2, 'ima_scale':'log', 'dark_rings': False}}

    Returns
    -------
    None
        updates the :class:`~matplotlib.figure.Figure` object

    Examples
    --------

    >>> from paos.core.parseConfig import parse_config
    >>> from paos.core.run import run
    >>> from paos.core.plot import simple_plot
    >>> pup_diameter, parameters, wavelengths, fields, opt_chains = parse_config('path/to/ini/file')
    >>> ret_val = run(pup_diameter, 1.0e-6 * wavelengths[0], parameters['grid size'],
    >>>              parameters['zoom'], fields[0], opt_chains[0])
    >>> from matplotlib import pyplot as plt
    >>> fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(8, 8))
    >>> key = list(ret_val.keys())[-1]  # plot at last optical surface
    >>> item = ret_val[key]
    >>> simple_plot(fig, ax, key=key, item=item, ima_scale='log')
    >>> plt.show()

    """
    ap, wx, wy = None, None, None
    logger.trace("plotting S{:02d}".format(key))

    if key in options.keys() and "pixel_units" in options[key].keys():
        assert isinstance(
            options[key]["pixel_units"], bool
        ), "pixel_units must be boolean"
        pixel_units = options[key]["pixel_units"]
    else:
        pixel_units = False

    if key in options.keys() and "dark_rings" in options[key].keys():
        assert isinstance(
            options[key]["dark_rings"], bool
        ), "dark_rings must be boolean"
        dark_rings = options[key]["dark_rings"]
    else:
        dark_rings = True

    if item["wz"] < 0.005:
        # Use microns
        scale = 1.0e6
        unit = "micron"
    else:
        # Use mm
        scale = 1.0e3
        unit = "mm"

    if "psf" in item.keys():
        ima = np.ma.masked_array(
            data=item["psf"], mask=item["amplitude"] <= 0.0
        )
    else:
        ima = np.ma.masked_array(
            data=item["amplitude"] ** 2, mask=item["amplitude"] <= 0.0
        )
    power = ima.sum()

    if key in options.keys() and "ima_scale" in options[key].keys():
        assert isinstance(
            options[key]["ima_scale"], str
        ), "ima_scale must be a str"
        assert options[key]["ima_scale"] in [
            "linear",
            "log",
        ], "ima_scale can be either linear or log"
        ima_scale = options[key]["ima_scale"]

    if ima_scale == "log":
        ima /= ima.max()
        im = axis.imshow(
            10 * np.ma.log10(ima),
            origin=origin,
            vmin=-20,
            vmax=0,
            cmap=plt.get_cmap(cmap),
        )
        cbar_label = "power/pix [db]"
    elif ima_scale == "linear":
        im = axis.imshow(ima, origin=origin, cmap=plt.get_cmap(cmap))
        cbar_label = "power/pix"
    else:
        logger.error("ima_scale shall be either log or linear")
        raise KeyError("ima_scale shall be either log or linear")

    cax = make_axes_locatable(axis).append_axes("right", size="5%", pad=0.05)
    cbar = fig.colorbar(im, cax=cax, orientation="vertical")
    cbar.set_label(cbar_label)

    if item["aperture"] is not None:
        x, y = item["aperture"].positions
        dx = 1.0 / scale if pixel_units else item["dx"]
        dy = 1.0 / scale if pixel_units else item["dy"]
        shapex = item["wfo"].shape[1]
        shapey = item["wfo"].shape[0]
        xy = scale * (x - shapex // 2) * dx, scale * (y - shapey // 2) * dy

        if hasattr(item["aperture"], "w") and hasattr(item["aperture"], "h"):
            w = item["aperture"].w
            h = item["aperture"].h
            wx = w * dx * scale
            wy = h * dy * scale
            xy = xy[0] - wx / 2, xy[1] - wy / 2
            ap = Rectangle(xy, wx, wy, ec="r", lw=5, fill=False)

        elif hasattr(item["aperture"], "a") and hasattr(item["aperture"], "b"):
            a = item["aperture"].a
            b = item["aperture"].b
            wx = 2 * a * dx * scale
            wy = 2 * b * dy * scale
            ap = Ellipse(xy, wx, wy, ec="r", lw=5, fill=False)

        axis.add_patch(ap)

    im.set_extent(np.array(item["extent"]) * scale)

    beam_radius = scale * item["wz"]
    airy_radius = 1.22 * scale * item["fratio"] * item["wl"]

    if np.isfinite(airy_radius) and dark_rings:
        for airy_scale in [1.22, 2.23, 3.24, 4.24, 5.24]:
            arad = airy_radius * airy_scale / 1.22
            width = (
                2.0 * arad / (scale * item["dx"])
                if pixel_units
                else 2.0 * arad
            )
            height = (
                2.0 * arad / (scale * item["dy"])
                if pixel_units
                else 2.0 * arad
            )
            aper = Ellipse(
                (0, 0),
                width=width,
                height=height,
                ec="k",
                lw=5,
                fill=False,
                alpha=0.5,
            )
            axis.add_patch(aper)

    if (
        beam_radius < airy_radius
        and (item["ABCDt"]() != np.eye(2)).all()
        and np.isfinite(airy_radius)
    ):
        plot_scale = 5.24 * airy_radius / 1.22  # 5th Airy null
    elif item["aperture"] is not None:
        plot_scale = max(wx / 2, wy / 2)
    else:
        plot_scale = beam_radius

    if key in options.keys():
        if "surface_zoom" in options[key].keys():
            plot_scale *= options[key]["surface_zoom"]
        elif "surface_scale" in options[key].keys():
            plot_scale = options[key]["surface_scale"]

    axis.set_title(
        rf"S{key:02d} | F#{item['fratio']:.2f} | w{scale * item['wz']:.2f}{unit:s} | "
        rf"$\lambda${1.0e6 * item['wl']:3.2f}$\mu$m | P{100 * power:2.0f}%"
    )

    if pixel_units:
        im.set_extent(
            np.array(item["extent"])
            / np.array([item["dx"], item["dx"], item["dy"], item["dy"]])
        )
        unit = "pixel"

    axis.set_xlabel(unit)
    axis.set_ylabel(unit)
    axis.set_xlim(-plot_scale, plot_scale)
    axis.set_ylim(-plot_scale, plot_scale)
    axis.grid()

    return


def plot_pop(retval, ima_scale="log", ncols=2, figname=None, options={}):
    """
    Given the POP simulation output dict, plots the squared amplitude of the
    wavefront at all the optical surfaces.

    Parameters
    ----------
    retval: dict
        simulation output dictionary
    ima_scale: str
        plot color map scale, can be either 'linear' or 'log'
    ncols: int
        number of columns for the subplots
    figname: str
        name of figure to save
    options: dict
        dict containing the options to display the plot: axis scale, axis unit, zoom scale and color scale.
        Examples:
        0) options={4: {'ima_scale':'linear'}}
        1) options={4: {'surface_scale':60, 'ima_scale':'linear'}}
        2) options={4: {'surface_scale':21, 'pixel_units':True, 'ima_scale':'linear'}}
        3) options={4: {'surface_zoom':2, 'ima_scale':'log'}}


    Returns
    -------
    out: None
        displays the plot output or stores it to the indicated plot path

    Examples
    --------

    >>> from paos.core.parseConfig import parse_config
    >>> from paos.core.run import run
    >>> from paos.core.plot import plot_pop
    >>> pup_diameter, parameters, wavelengths, fields, opt_chains = parse_config('path/to/ini/file')
    >>> ret_val = run(pup_diameter, 1.0e-6 * wavelengths[0], parameters['grid size'],
    >>>              parameters['zoom'], fields[0], opt_chains[0])
    >>> plot_pop(ret_val, ima_scale='log', ncols=3, figname='path/to/output/plot')

    """

    i, j = None, None

    n_subplots = len(retval)
    if ncols > n_subplots:
        ncols = n_subplots

    nrows = n_subplots // ncols
    if n_subplots % ncols:
        nrows += 1

    figsize = (8 * ncols, 6 * nrows)
    fig, ax = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize)
    fig.patch.set_facecolor("white")
    plt.subplots_adjust(hspace=0.3, wspace=0.5)

    for k, (key, item) in enumerate(retval.items()):

        if n_subplots == 1:
            axis = ax
        elif n_subplots <= ncols:
            axis = ax[k]
        else:
            i = k % ncols
            j = k // ncols
            axis = ax[j, i]

        simple_plot(
            fig=fig,
            axis=axis,
            key=key,
            item=item,
            ima_scale=ima_scale,
            options=options,
        )

        if n_subplots % ncols and k == n_subplots - 1:
            for col in range(i + 1, ncols):
                ax[j, col].set_visible(False)

    if figname is not None:
        fig.savefig(figname, bbox_inches="tight", dpi=150)
        plt.close()
    else:
        fig.tight_layout()
        plt.show()

    return


def plot_psf_xsec(
    fig,
    axis,
    key,
    item,
    ima_scale="linear",
    x_units="standard",
    surface_zoom=1,
):
    """
    Given the POP simulation output dict, plots the cross-sections of the squared amplitude of the
    wavefront at the given optical surface.

    Parameters
    ----------
    fig: :class:`~matplotlib.figure.Figure`
        instance of matplotlib figure artist
    key: int
        optical surface index
    item: dict
        optical surface dict
    ima_scale: str
        y axis scale, can be either 'linear' or 'log'
    x_units: str
        units for x axis. Default is 'standard', to have units of mm or microns.
        Can also be 'wave', i.e. :math:`\\textrm{Displacement} / (F_{num} \\lambda)`.
    surface_zoom: scalar
        Surface zoom: more increases the axis limits

    Returns
    -------
    out: None
        updates the `~matplotlib.figure.Figure` object
    Examples
    --------

    >>> import matplotlib.pyplot as plt
    >>> from paos.core.parseConfig import parse_config
    >>> from paos.core.run import run
    >>> from paos.core.plot import plot_psf_xsec
    >>> pup_diameter, parameters, wavelengths, fields, opt_chains = parse_config('/path/to/config/file')
    >>> wl_idx = 0  # choose the first wavelength
    >>> wavelength, opt_chain = wavelengths[wl_idx], opt_chains[wl_idx]
    >>> ret_val = run(pup_diameter, 1.0e-6 * wavelength, parameters['grid_size'], parameters['zoom'],
    >>>               fields[0], opt_chain)
    >>> key = list(ret_val.keys())[-1]  # plot at last optical surface
    >>> fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(16, 8))
    >>> plot_psf_xsec(fig=fig,axis=ax,key=key,item=ret_val[key],ima_scale='log',x_units='wave')

    """

    logger.trace("plotting S{:02d}".format(key))

    wx, wy = None, None

    if item["wz"] < 0.005:
        # Use microns
        scale = 1.0e6
        unit = "micron"
    else:
        # Use mm
        scale = 1.0e3
        unit = "mm"

    if "psf" in item.keys():
        ima = np.ma.masked_array(
            data=item["psf"], mask=item["amplitude"] <= 0.0
        )
    else:
        ima = np.ma.masked_array(
            data=item["amplitude"] ** 2, mask=item["amplitude"] <= 0.0
        )
    power = ima.sum()

    Npt = ima.shape[0]
    cross_idx = range(ima.shape[0])

    airy_radius = 1.22 * item["fratio"] * item["wl"] * scale
    dx, dy = item["dx"], item["dy"]

    if item["aperture"] is not None:

        if hasattr(item["aperture"], "w") and hasattr(item["aperture"], "h"):
            w = item["aperture"].w
            h = item["aperture"].h
            wx = w * dx * scale
            wy = h * dy * scale

        elif hasattr(item["aperture"], "a") and hasattr(item["aperture"], "b"):
            a = item["aperture"].a
            b = item["aperture"].b
            wx = 2 * a * dx * scale
            wy = 2 * b * dy * scale

    beam_radius = scale * item["wz"]
    if (
        beam_radius < airy_radius
        and (item["ABCDt"]() != np.eye(2)).all()
        and np.isfinite(airy_radius)
    ):
        plot_scale = 5.24 * airy_radius / 1.22  # 5th Airy null
    elif item["aperture"] is not None:
        plot_scale = max(wx / 2, wy / 2)
    else:
        plot_scale = beam_radius
    x_label = unit

    # Build the axis arrays
    airy_scale = 1.22 / airy_radius
    x_i = dx * scale * np.linspace(-Npt // 2, Npt // 2, Npt, endpoint=False)
    y_i = x_i * dy / dx

    if item["wz"] < 0.005:

        # Rescale the axis arrays
        x_i *= airy_scale
        y_i *= airy_scale

        # Airy 2D function, normalised to area 1
        xx, yy = np.meshgrid(x_i, y_i)
        r = np.pi * np.sqrt(xx**2 + yy**2) + 1.0e-30
        airy = (2.0 * j1(r) / r) ** 2
        normalization = 0.25 * np.pi * (x_i[1] - x_i[0]) * (y_i[1] - y_i[0])
        airy *= normalization

        if x_units == "standard":
            plot_scale = 5.24 / airy_scale
            x_i /= airy_scale
            y_i /= airy_scale

        if x_units == "wave":
            airy_scale = 1.0
            plot_scale = 5.24
            x_label = r"1 /F$\lambda$"

        # Plot Airy X and Y cross-sections
        axis.plot(x_i, airy[Npt // 2, ...], color="C4", label="Airy X-cut", linestyle="--")
        axis.plot(y_i, airy[..., Npt // 2], color="C5", label="Airy Y-cut", linestyle="--")
        axis.set_ylim(1.0e-10, airy.max())

        # plot vertical lines to mark the positions of the Airy dark rings and set the axis ticks
        x_ticks = (
            np.array(
                [
                    -5.24,
                    -4.24,
                    -3.24,
                    -2.23,
                    -1.22,
                    1.22,
                    2.23,
                    3.24,
                    4.24,
                    5.24,
                ]
            )
            / airy_scale
        )
        if surface_zoom <= 1.2:
            axis.vlines(x_ticks, *axis.get_ylim(), colors="k", lw=2, alpha=0.5)
            axis.set_xticks(list(x_ticks) + [0.0])

    # Plot ima X, Y, 45 deg and 135 deg cross-sections
    axis.plot(x_i, ima[Npt // 2, ...], "C0", label="X-cut")
    axis.plot(y_i, ima[..., Npt // 2], "C1", label="Y-cut")

    c = np.sqrt(x_i**2 + y_i**2) * np.sign(x_i)
    axis.plot(
        c,
        ima[cross_idx, cross_idx],
        "C2",
        label=r"45$^\circ$-cut",
    )
    axis.plot(
        c + 0.5 * np.sqrt((x_i[1] - x_i[0]) * (y_i[1] - y_i[0])),
        ima[cross_idx, cross_idx[::-1]],
        "C3",
        label=r"135$^\circ$-cut",
    )

    axis.set_title(
        r"S{:02d} | F#{:.2f} | w{:.2f}{:s} | $\lambda${:3.2f}$\mu$m | P{:2.0f}%".format(
            key,
            item["fratio"],
            scale * item["wz"],
            unit,
            1.0e6 * item["wl"],
            100 * power,
        )
    )

    axis.set_ylabel("Cross-sections")
    axis.set_xlabel(x_label)
    axis.get_xaxis().set_major_formatter(ticks.ScalarFormatter())
    axis.tick_params(labelrotation=45)
    axis.set_yscale(ima_scale)
    axis.set_xlim(-plot_scale * surface_zoom, plot_scale * surface_zoom)

    do_legend(axis=axis, ncol=3 if item["wz"] < 0.005 else 2)
    axis.grid()

    return


def plot_surface(key, retval, ima_scale, origin="lower", zoom=1, figname=None):
    """
    Given the optical surface key, the POP output dictionary and the image scale, plots the squared amplitude
    of the wavefront at the given surface (cross-sections and 2D plot)

    Parameters
    ----------
    key: int
        the key index associated to the optical surface
    retval: dict
        the POP output dictionary
    ima_scale: str
        the image scale. Can be either 'linear' or 'log'
    origin: str
        matplotlib plot origin. Defaults to 'lower'
    zoom: scalar
        the surface zoom factor: more increases the axis limits
    figname: str
        name of figure to save

    Returns
    -------
    out: :class:`~matplotlib.figure.Figure`
        the figure with the squared amplitude of the wavefront at the given surface

    """

    fig, axs = plt.subplots(nrows=1, ncols=2, figsize=(12, 6))
    # Xsec plot
    plot_psf_xsec(
        fig=fig,
        axis=axs[0],
        key=key,
        item=retval[key],
        ima_scale=ima_scale,
        surface_zoom=zoom,
    )
    # 2D plot
    simple_plot(
        fig=fig,
        axis=axs[1],
        key=key,
        item=retval[key],
        ima_scale=ima_scale,
        origin=origin,
        options={key: {"surface_zoom": zoom}},
    )
    fig.suptitle(axs[0].get_title(), fontsize=20)
    axs[0].set_title("X-sec view")
    axs[1].set_title("2D view")
    fig.tight_layout()

    if figname is not None:
        fig.savefig(figname, bbox_inches="tight", dpi=150)
        plt.close()

    return fig



import configparser
import os
import sys
from typing import List

import numpy as np

from paos import logger
from paos.classes.abcd import ABCD
from paos.util.material import Material


def getfloat(value):
    try:
        return np.float64(value)
    except:
        return np.nan


def parse_config(filename):
    """
    Parse an ini lens file

    Parameters
    ----------
    filename: string
        full path to ini file

    Returns
    -------
    pup_diameter: float
        pupil diameter in lens units
    parameters: dict
        Dictionary with parameters defined in the section 'general' of the ini file
    field: List
        list of fields
    wavelengths: List
        list of wavelengths
    opt_chain_list: List
        Each list entry is a dictionary of the optical surfaces in the .ini file, estimated at the given wavelength.
        (Relevant only for diffractive components)

    Examples
    --------

    >>> from paos.core.parseConfig import parse_config
    >>> pup_diameter, parameters, wavelengths, fields, opt_chains = parse_config('path/to/ini/file')

    """
    config = configparser.ConfigParser()
    filename = os.path.expanduser(filename)
    if not os.path.exists(filename) or not os.path.isfile(filename):
        logger.error(
            "Input file {} does not exist or is not a file. Quitting..".format(
                filename
            )
        )
        sys.exit()
    config.read(filename)

    # Parse parameters in section 'general'
    allowed_grid_size = [64, 128, 256, 512, 1024]
    allowed_zoom_val = [1, 2, 4, 8, 16]
    parameters = {
        "project": config["general"]["project"],
        "version": config["general"]["version"],
    }
    dtmp = config["general"].getint("grid_size")
    if dtmp not in allowed_grid_size:
        raise ValueError(
            "Grid size not allowed. Allowed values are", allowed_grid_size
        )
    parameters["grid_size"] = dtmp
    dtmp = config["general"].getint("zoom")
    if dtmp not in allowed_zoom_val:
        raise ValueError(
            "Zoom value not allowed. Allowed values are", allowed_zoom_val
        )
    elif dtmp == 1:
        logger.warning(
            "Zoom value is 1, i.e. the beam width occupies the whole of the grid width. "
            "This will result a PSF that is not Nyquist sampled."
        )
    parameters["zoom"] = dtmp

    lens_unit = config["general"].get("lens_unit", "")
    if lens_unit != "m":
        raise ValueError("Verify lens_unit=m in ini file")

    Tambient = config["general"].getfloat("Tambient")
    parameters["Tambient"] = Tambient
    Pambient = config["general"].getfloat("Pambient")
    parameters["Pambient"] = Pambient

    # Parse section 'wavelengths'
    wavelengths = []
    num = 1
    while True:
        _wl_ = config["wavelengths"].getfloat("w{:d}".format(num))
        if _wl_:
            wavelengths.append(_wl_)
        else:
            break

        num += 1

    # Parse section 'fields'
    fields = []
    num = 1

    while True:
        _fld_ = config["fields"].get("f{:d}".format(num))
        if _fld_:
            _fld_ = np.fromstring(_fld_, sep=",")
            _fld_ = np.tan(np.deg2rad(_fld_))
            fields.append({"us": _fld_[0], "ut": _fld_[1]})
        else:
            break

        num += 1

    # Parse sections 'lens_??'
    opt_chain_list = []
    pup_diameter = None  # input pupil pup_diameter
    for _wl_ in wavelengths:
        n1, n2 = None, None  # Refractive index
        glasslib = Material(_wl_, Tambient=Tambient, Pambient=Pambient)
        opt_chain = {}
        lens_num = 1
        while "lens_{:02d}".format(lens_num) in config:
            _data_ = {"num": lens_num}

            element = config["lens_{:02d}".format(lens_num)]
            lens_num += 1

            if element.getboolean("Ignore"):
                continue

            _data_["type"] = element.get("SurfaceType", None)
            _data_["R"] = getfloat(element.get("Radius", ""))
            _data_["T"] = getfloat(element.get("Thickness", ""))
            _data_["material"] = element.get("Material", None)

            _data_["is_stop"] = element.getboolean("Stop", False)
            _data_["save"] = element.getboolean("Save", False)
            _data_["name"] = element.get("Comment", "")

            if _data_["type"] == "INIT":
                n1 = 1.0
                aperture = element.get("aperture", "").split(",")
                aperture_shape, aperture_type = aperture[0].split()
                if (
                    aperture_shape == "elliptical"
                    and aperture_type == "aperture"
                ):
                    xpup = getfloat(aperture[2])
                    ypup = getfloat(aperture[3])
                    pup_diameter = 2.0 * max(xpup, ypup)

                continue

            if n1 is None or pup_diameter is None:
                # logger.error('INIT is not the first surface in Lens Data.')
                raise ValueError("INIT is not the first surface in Lens Data.")

            if _data_["type"] == "Zernike":
                thickness = 0.0
                curvature = 0.0
                n2 = n1
                wave = 1.0e-6 * getfloat(element.get("Par1", ""))
                _data_["Zordering"] = element.get("Par2", "").lower()
                _data_["Znormalize"] = element.getboolean("Par3")
                _data_["Zradius"] = getfloat(element.get("Par4", ""))
                _data_["Zorigin"] = element.get("Par5", "x")

                _data_["Zindex"] = np.fromstring(
                    element.get("Zindex", ""), sep=",", dtype=np.int64
                )
                _data_["Z"] = (
                    np.fromstring(
                        element.get("Z", ""), sep=",", dtype=np.float64
                    )
                    * wave
                )

                _data_["ABCDt"] = ABCD(
                    thickness=thickness,
                    curvature=curvature,
                    n1=n1,
                    n2=n2,
                    M=1.0,
                )
                _data_["ABCDs"] = ABCD(
                    thickness=thickness,
                    curvature=curvature,
                    n1=n1,
                    n2=n2,
                    M=1.0,
                )

            elif _data_["type"] == "Coordinate Break":
                thickness = _data_["T"] if np.isfinite(_data_["T"]) else 0.0
                curvature = 0.0
                n2 = n1
                _data_["xdec"] = getfloat(element.get("Par1", ""))
                _data_["ydec"] = getfloat(element.get("Par2", ""))
                _data_["xrot"] = getfloat(element.get("Par3", ""))
                _data_["yrot"] = getfloat(element.get("Par4", ""))

                _data_["ABCDt"] = ABCD(
                    thickness=thickness,
                    curvature=curvature,
                    n1=n1,
                    n2=n2,
                    M=1.0,
                )
                _data_["ABCDs"] = ABCD(
                    thickness=thickness,
                    curvature=curvature,
                    n1=n1,
                    n2=n2,
                    M=1.0,
                )

            elif _data_["type"] == "Paraxial Lens":
                focal_length = getfloat(element.get("Par1", ""))
                thickness = _data_["T"] if np.isfinite(_data_["T"]) else 0.0
                curvature = (
                    1 / focal_length if np.isfinite(focal_length) else 0.0
                )
                n2 = n1
                aperture = element.get("aperture", "")
                if aperture:
                    aperture = aperture.split(",")
                    aperture_shape, aperture_type = aperture[0].split()
                    _data_["aperture"] = {
                        "shape": aperture_shape,
                        "type": aperture_type,
                        "xrad": getfloat(aperture[1]),
                        "yrad": getfloat(aperture[2]),
                        "xc": getfloat(aperture[3]),
                        "yc": getfloat(aperture[4]),
                    }
                _data_["ABCDt"] = ABCD(
                    thickness=thickness,
                    curvature=curvature,
                    n1=n1,
                    n2=n2,
                    M=1.0,
                )
                _data_["ABCDs"] = ABCD(
                    thickness=thickness,
                    curvature=curvature,
                    n1=n1,
                    n2=n2,
                    M=1.0,
                )

            elif _data_["type"] == "ABCD":
                thickness = _data_["T"] if np.isfinite(_data_["T"]) else 0.0
                Ax = getfloat(element.get("Par1", ""))
                Bx = getfloat(element.get("Par2", ""))
                Cx = getfloat(element.get("Par3", ""))
                Dx = getfloat(element.get("Par4", ""))
                Ay = getfloat(element.get("Par5", ""))
                By = getfloat(element.get("Par6", ""))
                Cy = getfloat(element.get("Par7", ""))
                Dy = getfloat(element.get("Par8", ""))
                ABCDs = ABCD(
                    thickness=thickness, curvature=0.0, n1=n1, n2=n1, M=1.0
                )
                ABCDt = ABCD(
                    thickness=thickness, curvature=0.0, n1=n1, n2=n1, M=1.0
                )
                _ABCDs = np.array([[Ax, Bx], [Cx, Dx]])
                _ABCDt = np.array([[Ay, By], [Cy, Dy]])
                ABCDs.ABCD = ABCDs() @ _ABCDs
                ABCDt.ABCD = ABCDt() @ _ABCDt
                aperture = element.get("aperture", "")
                if aperture:
                    aperture = aperture.split(",")
                    aperture_shape, aperture_type = aperture[0].split()
                    _data_["aperture"] = {
                        "shape": aperture_shape,
                        "type": aperture_type,
                        "xrad": getfloat(aperture[1]),
                        "yrad": getfloat(aperture[2]),
                        "xc": getfloat(aperture[3]),
                        "yc": getfloat(aperture[4]),
                    }
                _data_["ABCDt"] = ABCDt
                _data_["ABCDs"] = ABCDs

            elif _data_["type"] == "Standard":
                thickness = _data_["T"] if np.isfinite(_data_["T"]) else 0.0
                curvature = (
                    1 / _data_["R"] if np.isfinite(_data_["R"]) else 0.0
                )
                aperture = element.get("aperture", "")
                if aperture:
                    aperture = aperture.split(",")
                    aperture_shape, aperture_type = aperture[0].split()
                    _data_["aperture"] = {
                        "shape": aperture_shape,
                        "type": aperture_type,
                        "xrad": getfloat(aperture[1]),
                        "yrad": getfloat(aperture[2]),
                        "xc": getfloat(aperture[3]),
                        "yc": getfloat(aperture[4]),
                    }

                if _data_["material"] == "MIRROR":
                    n2 = -n1
                elif _data_["material"] in glasslib.materials.keys():
                    n2 = glasslib.nmat(_data_["material"])[1] * np.sign(n1)
                else:
                    n2 = 1.0 * np.sign(n1)

                _data_["ABCDt"] = ABCD(
                    thickness=thickness,
                    curvature=curvature,
                    n1=n1,
                    n2=n2,
                    M=1.0,
                )
                _data_["ABCDs"] = ABCD(
                    thickness=thickness,
                    curvature=curvature,
                    n1=n1,
                    n2=n2,
                    M=1.0,
                )

            else:
                logger.error(
                    "Surface Type not recognised: {:s}".format(
                        str(_data_["type"])
                    )
                )
                raise ValueError(
                    "Surface Type not recognised: {:s}".format(
                        str(_data_["type"])
                    )
                )

            opt_chain[_data_["num"]] = _data_
            n1 = n2

        opt_chain_list.append(opt_chain)

    return pup_diameter, parameters, wavelengths, fields, opt_chain_list



from math import factorial as fac

import numpy as np
from scipy.special import eval_jacobi as jacobi


class Zernike:
    """
    Generates Zernike polynomials

    Parameters
    ----------
    N : integer
        Number of polynomials to generate in a sequence following the defined 'ordering'
    rho : array like
        the radial coordinate normalised to the interval [0, 1]
    phi : array like
        Azimuthal coordinate in radians. Has same shape as rho.
    ordering : string
        Can be either ANSI ordering (ordering='ansi', this is the default), or Noll ordering
        (ordering='noll'), or Fringe ordering (ordering='fringe'), or Standard (Born&Wolf) ordering (ordering='standard')
    normalize : bool
        Set to True generates ortho-normal polynomials. Set to False generates orthogonal polynomials
        as described in `Laksminarayan & Fleck, Journal of Modern Optics (2011) <https://doi.org/10.1080/09500340.2011.633763>`_.
        The radial polynomial is estimated using the Jacobi polynomial expression as in their Equation in Equation 14.


    Returns
    -------
    out : masked array
        An instance of Zernike.

    Example
    -------
    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> x = np.linspace(-1.0, 1.0, 1024)
    >>> xx, yy = np.meshgrid(x, x)
    >>> rho = np.sqrt(xx**2 + yy**2)
    >>> phi = np.arctan2(yy, xx)
    >>> zernike = Zernike(36, rho, phi, ordering='noll', normalize=True)
    >>> zer = zernike() # zer contains a list of polynomials, noll-ordered

    >>> # Plot the defocus zernike polynomial
    >>> plt.imshow(zer[3])
    >>> plt.show()

    >>> # Plot the defocus zernike polynomial
    >>> plt.imshow(zernike(3))
    >>> plt.show()

    Note
    ----
    In the example, the polar angle is counted counter-clockwise positive from the
    x axis. To have a polar angle that is clockwise positive from the y axis
    (as in figure 2 of `Laksminarayan & Fleck, Journal of Modern Optics (2011) <https://doi.org/10.1080/09500340.2011.633763>`_) use

    >>> phi = 0.5*np.pi - np.arctan2(yy, xx)

    """

    def __init__(self, N, rho, phi, ordering="ansi", normalize=False):

        assert ordering in (
            "ansi",
            "noll",
            "fringe",
            "standard",
        ), "Unrecognised ordering scheme."
        assert N > 0, "N shall be a positive integer"

        self.ordering = ordering
        self.N = N
        self.m, self.n = self.j2mn(N, ordering)

        if normalize:
            self.norm = [
                np.sqrt(n + 1) if m == 0 else np.sqrt(2.0 * (n + 1))
                for m, n in zip(self.m, self.n)
            ]
        else:
            self.norm = np.ones(self.N, dtype=np.float64)

        mask = rho > 1.0
        if isinstance(rho, np.ma.MaskedArray):
            rho.mask |= mask
        else:
            rho = np.ma.MaskedArray(data=rho, mask=mask, fill_value=0.0)

        Z = {}
        for n in range(max(self.n) + 1):
            Z[n] = {}
            for m in range(-n, 1, 2):
                Z[n][m] = data = self.__ZradJacobi__(m, n, rho)
                Z[n][-m] = Z[n][m].view()

        self.Zrad = [Z[n][m].view() for m, n in zip(self.m, self.n)]

        Z = {0: np.ones_like(phi)}
        for m in range(1, self.m.max() + 1):
            Z[m] = np.cos(m * phi)
            Z[-m] = np.sin(m * phi)
        self.Zphi = [Z[m].view() for m in self.m]

        self.Z = np.ma.MaskedArray(
            [
                self.norm[k] * self.Zrad[k] * self.Zphi[k]
                for k in range(self.N)
            ],
            fill_value=0.0,
        )

    def __call__(self, j=None):
        """
        Parameters
        ----------
        j : integer
            Polynomial to return. If set to None, returns all polynomial requested at
            instantiation

        Returns
        -------
        out : masked array
          if j is set to None, the output is a masked array where the first dimension
          has the size of the number of polynomials requested.
          When j is set to an integer, returns the j-th polynomial as a masked array.
        """
        if j is None:
            return self.Z
        else:
            return self.Z[j]

    @staticmethod
    def j2mn(N, ordering):
        """
        Convert index j into azimuthal number, m, and radial number, n
        for the first N Zernikes

        Parameters
        ----------
        N: integer
            Number of polynomials (starting from Piston)
        ordering: string
            can take values 'ansi', 'standard', 'noll', 'fringe'

        Returns
        -------
        m, n: array

        """
        j = np.arange(N, dtype=int)

        if ordering == "ansi":
            n = np.ceil((-3.0 + np.sqrt(9.0 + 8.0 * j)) / 2.0).astype(int)
            m = 2 * j - n * (n + 2)
        elif ordering == "standard":
            n = np.ceil((-3.0 + np.sqrt(9.0 + 8.0 * j)) / 2.0).astype(int)
            m = -2 * j + n * (n + 2)
        elif ordering == "noll":
            index = j + 1
            n = ((0.5 * (np.sqrt(8 * index - 7) - 3)) + 1).astype(int)
            cn = n * (n + 1) / 2 + 1
            m = np.empty(N, dtype=int)
            idx = n % 2 == 0
            m[idx] = (index[idx] - cn[idx] + 1) // 2 * 2
            m[~idx] = (index[~idx] - cn[~idx]) // 2 * 2 + 1
            m = (-1) ** (index % 2) * m
        elif ordering == "fringe":
            index = j + 1
            m_n = 2 * (np.ceil(np.sqrt(index)) - 1)
            g_s = (m_n / 2) ** 2 + 1
            n = m_n / 2 + np.floor((index - g_s) / 2)
            m = (m_n - n) * (1 - np.mod(index - g_s, 2) * 2)
            return m.astype(int), n.astype(int)
        else:
            raise NameError("Ordering not supported.")

        return m, n

    @staticmethod
    def mn2j(m, n, ordering):
        """
        Convert radial and azimuthal numbers, respectively n and m, into index j
        """
        if ordering == "ansi":
            return (n * (n + 2) + m) // 2
        elif ordering == "standard":
            return (n * (n + 2) - m) // 2
        elif ordering == "fringe":
            a = (1 + (n + np.abs(m)) / 2) ** 2
            b = 2 * np.abs(m)
            c = (1 + np.sign(m)) / 2
            return (a - b - c).astype(int) + 1
        elif ordering == "noll":
            _p = np.zeros(n.size, dtype=np.int64)
            for idx, (_m, _n) in enumerate(zip(m, n)):
                if _m > 0.0 and (_n % 4 in [0, 1]):
                    _p[idx] = 0
                elif _m < 0.0 and (_n % 4 in [2, 3]):
                    _p[idx] = 0
                elif _m >= 0.0 and (_n % 4 in [2, 3]):
                    _p[idx] = 1
                elif _m <= 0.0 and (_n % 4 in [0, 1]):
                    _p[idx] = 1
                else:
                    raise ValueError("Invalid (m,n) in Noll indexing.")
            return (n * (n + 1) / 2 + np.abs(m) + _p).astype(np.int64)
        else:
            raise NameError("Ordering not supported.")

    @staticmethod
    def __ZradJacobi__(m, n, rho):
        """
        Computes the radial Zernike polynomial

        Parameters
        ----------
        m : integer
          azimuthal number
        n : integer
          radian number
        rho : array like
          Pupil semi-diameter normalised radial coordinates

        Returns
        -------
        R_mn : array like
          the radial Zernike polynomial with shape identical to rho

        """

        m = np.abs(m)

        if n < 0:
            raise ValueError(
                "Invalid parameter: n={:d} should be > 0".format(n)
            )
        if m > n:
            raise ValueError(
                "Invalid parameter: n={:d} should be larger than m={:d}".format(
                    n, m
                )
            )
        if (n - m) % 2:
            raise ValueError(
                "Invalid parameter: n-m={:d} should be a positive even number.".format(
                    n - m
                )
            )

        jpoly = jacobi((n - m) // 2, m, 0.0, (1.0 - 2.0 * rho**2))

        return (-1) ** ((n - m) // 2) * rho**m * jpoly

    @staticmethod
    def __ZradFactorial__(m, n, rho):
        """
        CURRENTLY NOT USED
        Computes the radial Zernike polynomial

        Parameters
        ----------
        m : integer
          azimuthal number
        n : integer
          radian number
        rho : array like
          Pupil semi-diameter normalised radial coordinates

        Returns
        -------
        R_mn : array like
          the radial Zernike polynomial with shape identical to rho

        """
        m = np.abs(m)

        if n < 0:
            raise ValueError(
                "Invalid parameter: n={:d} should be > 0".format(n)
            )
        if m > n:
            raise ValueError(
                "Invalid parameter: n={:d} should be larger than m={:d}".format(
                    n, m
                )
            )
        if (n - m) % 2:
            raise ValueError(
                "Invalid parameter: n-m={:d} should be a positive even number.".format(
                    n - m
                )
            )

        pre_fac = (
            lambda k: (-1.0) ** k
            * fac(n - k)
            / (fac(k) * fac((n + m) // 2 - k) * fac((n - m) // 2 - k))
        )

        return sum(
            pre_fac(k) * rho ** (n - 2 * k) for k in range((n - m) // 2 + 1)
        )

    def cov(self):
        """
        Computes the covariance matrix M defined as

        >>> M[i, j] = np.mean(Z[i, ...]*Z[j, ...])

        When a pupil is defined as :math:`\\Phi = \\sum c[k] Z[k, ...]`, the pupil RMS can be calculated as

        >>> RMS = np.sqrt( np.dot(c, np.dot(M, c)) )

        This works also on a non-circular pupil, provided that the polynomials are masked over the pupil.

        Returns
        -------
        M : array
          the covariance matrix
        """
        cov = np.empty((self.Z.shape[0], self.Z.shape[0]))
        for i in range(self.Z.shape[0]):
            for j in range(i, self.Z.shape[0]):
                cov[i, j] = cov[j, i] = np.ma.mean(self.Z[i] * self.Z[j])

        cov[cov < 1e-10] = 0.0

        return cov



import configparser
import gc
import io
import itertools
import os
import re
import sys
from tkinter import Tk
from typing import List

from matplotlib import pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasAgg
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from PySimpleGUI import Canvas
from PySimpleGUI import Checkbox
from PySimpleGUI import clipboard_set
from PySimpleGUI import Column
from PySimpleGUI import Frame
from PySimpleGUI import InputText
from PySimpleGUI import pin
from PySimpleGUI import popup_get_file
from PySimpleGUI import popup_quick_message
from PySimpleGUI import ProgressBar
from PySimpleGUI import Text
from PySimpleGUI import Window

from paos import logger
from paos import save_datacube
from paos.core.plot import plot_psf_xsec
from paos.core.plot import simple_plot


class SimpleGui:
    """
    Base class for the Graphical User Interface (GUI) for ``PAOS``, built using the publicly available library PySimpleGUI
    """

    def __init__(self):
        """
        Initializes the GUI.
        This includes instantiating the global variables, defining the symbols to use
        and defining the GUI Menu (principal and right-click)
        """

        # ------ Instantiate global variables ------ #
        self.window = None
        self.config = None
        self.temporary_config = None
        self.font_titles = ("Helvetica", 20)
        self.font_subtitles = ("Helvetica", 18)
        self.font_small = ("Courier New", 10)
        self.font_underlined = ("Courier New underline", 16)

        # ------ Menu Definition ------ #
        self.menu_def = [
            [
                "&File",
                ["&Open", "&Save", "&Save As", "Global Settings", "&Exit"],
            ],
            ["&Help", "&About"],
        ]

        # ------ Right Click Menu Definition ------ #
        self.right_click_menu_def = ["", ["Nothing", "Version", "Exit"]]

        # ------ Symbols Definition ------ #
        self.triangle_right = "▶"
        self.triangle_down = "▼"
        self.symbol_disabled = "◯"
        self.symbol_state = "●"

        # ------ Quick Message Definition ------ #
        w, h = Window.get_screen_size()
        popup_quick_message(
            "Hang on for a moment, this will take a bit to create....",
            auto_close=True,
            non_blocking=True,
            keep_on_top=True,
            auto_close_duration=2,
            location=(int(0.4 * w), int(0.1 * h)),
        )

    @staticmethod
    def add_heading(headings, size=(24, 2)):
        """
        Given a list of header names and a tuple for the size, returns a chained list of Text widgets with the specified
        size where the first element is returned to prettify spacing

        Parameters
        ----------
        headings: list
            a list of header names
        size: tuple
            a tuple for the widget sizes defined as (width, height)

        Returns
        -------
        out: List[Text]
            a chained list of Text widgets

        """
        headings_list = []
        for i, head in enumerate(headings):
            headings_list.append(Text(head, key=head, size=size))
        return list(itertools.chain([Text(" " * 8)], headings_list))

    @staticmethod
    def collapse_frame(title, layout, key):
        """
        Helper function that creates a Frame that can be later made hidden, thus appearing "collapsed"

        Parameters
        ----------
        title: str
            the Frame title
        layout: List[List[Element]]
            the layout for the section
        key: str
            key used to make this section visible / invisible

        Returns
        -------
        out: Frame
            A pinned Frame that can be placed directly into your layout

        """
        return pin(Frame(title=title, layout=layout, key=key, visible=False))

    @staticmethod
    def collapse_column(layout, key):
        """
        Helper function that creates a Column that can be later made hidden, thus appearing "collapsed"

        Parameters
        ----------
        layout: List[List[Element]]
            the layout for the section
        key: str
            key used to make this section visible / invisible

        Returns
        -------
        out: Column
            A pinned Column that can be placed directly into your layout

        """
        return pin(Column(layout=layout, key=key, visible=False))

    @staticmethod
    def update_column_scrollbar(window, col_key):
        """
        Given the current GUI window and the current Column key, updates the column scrollbar if the Column has changed

        Parameters
        ----------
        window: Window
            the current GUI window
        col_key: str
            the current Column key

        Returns
        -------
        out: None
            Updates the column scrollbar

        """
        window.refresh()
        window[col_key].contents_changed()
        return

    @staticmethod
    def get_clipboard_text():
        """
        Returns a copy of the local clipboard content (e.g. an Excel column)

        Returns
        -------
        out: List[str]
            the local copy of the clipboard's content
        """
        root = Tk()
        root.withdraw()
        text = root.clipboard_get().replace("\n", ",").split(",")
        text = text if text[-1] != "" else text[:-1]
        return text

    @staticmethod
    def to_configparser(dictionary):
        """
        Given a dictionary, it converts it into a :class:`~configparser.ConfigParser` object

        Parameters
        ----------
        dictionary: dict
            input dictionary to be converted

        Returns
        -------
        out: :class:`~configparser.ConfigParser`
        """

        config = configparser.ConfigParser()

        for key, item in dictionary.items():
            if isinstance(item, dict):
                config.add_section(key)
                for subkey, subitem in item.items():
                    if subitem is not None and subitem != "":
                        if isinstance(subitem, str):
                            pass
                        elif isinstance(subitem, (float, bool)):
                            subitem = str(subitem)
                        elif isinstance(subitem, (tuple, list)):
                            subitem = ",".join(subitem)
                        else:
                            raise NotImplementedError("item type not supported")
                    config.set(key, subkey, subitem)

        return config

    def to_hdf5(self, retval_list, groups, keys_to_keep=None):
        """
        Given the POP output dictionary list, the names for each simulation (saving groups) and the keys to store,
        opens a popup to get the output file name and then dumps the simulation outputs to a hdf5 file.

        Parameters
        ----------
        retval_list: List of dict
            list of simulation output dictionaries
        groups: List of strings
            names to associate to each simulation in the output
        keys_to_keep: List of strings
            dictionary keys to store

        Returns
        -------
        out: None
            Dumps the simulation output to the indicated hdf5 file

        """

        if keys_to_keep is None:
            keys_to_keep = ["amplitude", "dx", "dy", "wl"]

        if not retval_list:
            logger.error("Run POP first")
            return

        # Get the file path to save to
        filename = popup_get_file(
            "Choose file (HDF5) to save to",
            default_path=self.passvalue["output"],
            default_extension=".h5",
            save_as=True,
            keep_on_top=True,
        )

        if filename is None:
            logger.warning("Pressed Cancel. Continuing...")
            return

        if not filename.endswith((".HDF5", ".hdf5", ".H5", ".h5")):
            logger.debug("Saving file format not provided. Defaulting to .h5")
            filename = "".join([filename, ".h5"])

        # Save the POP output to the specified .hdf5 file
        tags = list(map(str, groups))
        save_datacube(
            retval_list,
            filename,
            tags,
            keys_to_keep=keys_to_keep,
            overwrite=True,
        )
        return

    @staticmethod
    def to_txt(text_list):
        """
        Given a list of strings, opens a popup to get the output file name and then dumps the text ordered in rows
        to the output text file

        Parameters
        ----------
        text_list: list of strings

        Returns
        -------
        out: None
            writes the input text list to a text file.
            Used to save the output of the diagnostic raytrace

        """

        if text_list == "":
            logger.error("Perform raytrace first")
            return

        # Get the file path to save to
        filename = popup_get_file(
            "Choose file (TXT) to save to", save_as=True, keep_on_top=True
        )

        if filename is None:
            logger.warning("Pressed Cancel. Continuing...")
            return

        if not filename.endswith((".TXT", ".txt")):
            logger.debug("Saving file format not provided. Defaulting to .txt")
            filename = "".join([filename, ".txt"])

        # Save the text list to the specified .txt file
        with open(filename, "w") as f:
            f.write("\n".join(text_list))
        return

    @staticmethod
    def draw_figure(figure, canvas):
        """
        CURRENTLY NOT USED
        Given a Canvas and a figure, it draws the figure onto the Canvas

        Parameters
        ----------
        figure: :class:`~matplotlib.figure.Figure`
            the figure to be drawn
        canvas: Canvas
            the canvas onto which to draw the figure

        Returns
        -------
        out: FigureCanvasTkAgg
            the Tkinter widget to draw a :class:`~matplotlib.figure.Figure` onto a Canvas
        """
        figure_canvas_agg = FigureCanvasTkAgg(figure, canvas)
        figure_canvas_agg.draw()
        figure_canvas_agg.get_tk_widget().pack(side="top", fill="both", expand=1)
        return figure_canvas_agg

    @staticmethod
    def draw_image(figure, element):
        """
        Draws the previously created "figure" in the supplied Image Element

        Parameters
        ----------
        figure: :class:`~matplotlib.figure.Figure`
            a Matplotlib figure
        element: Image
            an Image element

        Returns
        -------
        out: Canvas
            The figure canvas

        """

        if figure is None:
            logger.error("Plot first")
            return

        plt.close("all")  # erases previously drawn plots
        canvas = FigureCanvasAgg(figure)
        buf = io.BytesIO()
        canvas.print_figure(buf, format="png")
        if buf is not None:
            buf.seek(0)
            element.update(data=buf.read())
            element.update(visible=True)
            return canvas
        else:
            return None

    @staticmethod
    def clear_image(element):
        """
        Given an Image widget, it clears it

        Parameters
        ----------
        element: Image
            the Image widget to clear

        Returns
        -------
        out: None
            clears an Image widget
        """
        logger.debug("Clearing image")
        element.update()
        element.update(visible=False)

        return

    @staticmethod
    def save_figure(figure, filename=None):
        """
        Given a matplotlib figure and a file path (optional), saves the figure to the file path

        Parameters
        ----------
        figure: :class:`~matplotlib.figure.Figure`
            the matplotlib figure to save
        filename: str
            the saving file path

        Returns
        -------
        out: None
            Saves the matplotlib figure

        """
        if figure is None:
            logger.error("Create plot first")
            return

        if filename is None:
            # Get the file path to save to
            filename = popup_get_file(
                "Choose file (PNG, JPG) to save to",
                save_as=True,
                keep_on_top=True,
            )

            if filename is None:
                logger.warning("Pressed Cancel. Continuing...")
                return

            if not filename.endswith((".PNG", ".png", ".JPG", ".jpg")):
                logger.debug("Saving file format not provided. Defaulting to .png")
                filename = "".join([filename, ".png"])

        # Save the plot to the specified .png or .jpg file
        figure.savefig(filename, bbox_inches="tight", dpi=150)

        logger.debug(f"Saved figure to {filename}")

        return

    @staticmethod
    def reset_progress_bar(progress_bar):
        """
        Given a progress bar element, it resets it

        Parameters
        ----------
        progress_bar: ProgressBar
            the progress bar element to reset

        Returns
        -------
        out: ProgressBar
            resets the progress bar element
        """
        progress_bar.metadata = 0
        progress_bar.update_bar(progress_bar.metadata)
        return progress_bar

    @staticmethod
    def move_with_arrow_keys(window, event, values, elem_key, max_rows, max_cols):
        """
        Given the current GUI window, the latest event, the dictionary containing the window values, the dictionary key
        for the cell with focus and the maximum sizes for the current table editor, this method sets the focus on
        the new cell (if the current cell changed).

        Parameters
        ----------
        window: Window
            the current GUI window
        event: str
            the latest GUI event
        values: dict
            the dictionary containing the window values
        elem_key: str
            the dictionary key for the cell with focus
        max_rows: int
            the number of rows for the current table editor
        max_cols: int
            the number of columns for the current table editor

        Returns
        -------
        out: int
            The row number corresponding to where the current focus is
        """
        current_cell = re.findall("[0-9]+", elem_key.partition("_")[-1])
        r, c = tuple(map(int, current_cell))

        if event.startswith("Down"):
            r = r + 1 * (r < max_rows)
        elif event.startswith("Left"):
            c = c - 1 * (c > 0)
        elif event.startswith("Right"):
            c = c + 1 * (c < max_cols - 1)
        elif event.startswith("Up"):
            r = r - 1 * (r > 0) if r > 1 else 1

        # if the current cell changed, set focus on new cell
        if current_cell != (r, c):
            for key in values.keys():
                if key.endswith(f"({r},{c})"):
                    window[key].set_focus()  # set the focus on the element moved to
                    window[key].update()
        return r

    @staticmethod
    def copy_to_clipboard(dictionary):
        """
        Saves the relevant data from the GUI configuration tabs into a dictionary, then copies it to the local clipboard

        Returns
        -------
        out: None
            copies the data to the local clipboard
        """
        clipboard_set(str(dictionary))
        return

    @staticmethod
    def chain_widgets(row, input_list, prefix, disabled_list=None):
        """
        Given the row in the GUI editor, an input list and the key prefix, returns a list of widgets

        Parameters
        ----------
        row: int
            the current row in the GUI editor
        input_list: List
            the item list to insert in the cell widgets
        prefix: str
            the key prefix
        disabled_list: list[bool]
            list with options to disable widgets

        Returns
        -------
        out: List[Text, List[Checkbox or Input or Column or Combo]]
            list of widgets
        """
        if disabled_list is None:
            disabled_list = []
        row_widget = [Text(row, size=(6, 1), key=f"row idx {row}")]
        keys = [f"{prefix}_({row},{i})" for i in range(len(input_list))]

        if not disabled_list:
            disabled_list = [False for _ in input_list]

        return list(
            itertools.chain(
                row_widget,
                [
                    InputText(
                        default_text=value,
                        key=key,
                        size=(24, 2),
                        disabled=disabled,
                    )
                    for value, key, disabled in zip(input_list, keys, disabled_list)
                ],
            )
        )

    def make_visible(self, event, visible, key):
        """
        Given the current event, a boolean and the widget key, it makes the widget visible/invisible

        Parameters
        ----------
        event: str
            the current event
        visible: bool
            if True, the widget becomes visible. If False, it becomes invisible
        key: str
            the widget key

        Returns
        -------
        out: bool
            the updated visibile parameter of the widget

        """
        visible = not visible
        self.window[key].update(visible=visible)
        # Update the triangle symbol
        self.window[event].update(
            self.triangle_down if visible else self.triangle_right
        )

        return visible

    def close_window(self):
        """
        Deletes the temporary .ini file (if present), any unsaved outputs and then closes the GUI window

        Returns
        -------
        out: None
            closes the GUI window and cleans up the memory
        """

        # ------ Remove temporary config ------ #
        if self.temporary_config is not None:
            if not os.path.exists(self.temporary_config) or not os.path.isfile(
                self.temporary_config
            ):
                logger.error(
                    f"Input temporary file {self.temporary_config} does not exist or is not a file. "
                    f"Quitting.."
                )
                sys.exit()
            logger.info("Removing temporary .ini configuration file")
            os.remove(self.temporary_config)

        # ------ Close plots ------ #
        plt.close("all")

        # ------ Close Window ------ #
        self.window.close()
        del self.window

        # ------ Collect garbage ------ #
        _ = gc.collect()

        return

    @staticmethod
    def sort_column(window, values, col_key):
        """
        Given a GUI window, the window contents (values) and the Column key, it sorts the Column elements in
        increasing order

        Parameters
        ----------
        window: Window
            the GUI window
        values: dict
            the dictionary with the Window contents
        col_key: str
            the Column key

        Returns
        -------
        out: None
            Sorts the Column elements in ascending order

        """
        col_values = []
        keys = []
        for key, item in values.items():
            if key.startswith(col_key):
                col_values.append(item)
                keys.append(key)

        if "" in col_values:
            logger.warning("Invalid wavelength in column. Continuing..")
            return

        col_values = list(map(float, col_values))
        col_values.sort()
        col_values = list(map(str, col_values))

        for key, value in zip(keys, col_values):
            window[key].update(value)

        return



# This code is inspired by the logger developed for ExoSim2.0 which is in turn
# inspired by the logger developed for TauREx3.1
# Therefore, we attach here the ExoSim2.0 and the TauREx3.1 license:
#
# BSD 3-Clause License
#
# Copyright (c) 2020, arielmission-space
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS”
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#
# BSD 3-Clause License
#
# Copyright (c) 2019, Ahmed F. Al-Refaie, Quentin Changeat, Ingo Waldmann, Giovanna Tinetti
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the names of the copyright holders nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import logging

from decorator import decorator

from paos import __pkg_name__

__all__ = ["Logger"]

root_logger = logging.getLogger(__pkg_name__)
root_logger.propagate = False


class CustomFormatter(logging.Formatter):
    """Custom formatter for logging.
    Different colors propagate to the terminal depending on the level of the log.
    """

    grey = "\x1b[38;20m"
    green = "\x1b[32;20m"
    brown = "\x1b[0;33m"
    blue = "\x1b[34;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

    FORMATS = {
        10: grey + format + reset,
        15: grey + format + reset,
        20: green + format + reset,
        30: yellow + format + reset,
        40: red + format + reset,
        50: bold_red + format + reset,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


# formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
ch = logging.StreamHandler()
ch.setFormatter(CustomFormatter())
ch.setLevel(logging.DEBUG)
root_logger.addHandler(ch)
root_logger.setLevel(logging.DEBUG)


class Logger:
    """
    *Abstract class*

    Standard logging using logger library.
    It's an abstract class to be inherited to load its methods for logging.
    It define the logger name at the initialization, and then provides the logging methods.
    """

    def __init__(self):
        self._logger = None
        self._log_name = None
        self.set_log_name()

    def set_log_name(self):
        """
        Produces the logger name and store it inside the class.
        The logger name is the name of the class that inherits this Logger class.
        """
        self._log_name = "{}.{}".format(__pkg_name__, self.__class__.__name__)
        self._logger = logging.getLogger(
            "{}.{}".format(__pkg_name__, self.__class__.__name__)
        )

    def info(self, message, *args, **kwargs):
        """
        Produces INFO level log
        See :class:`logging.Logger`
        """
        self._logger.info(message, *args, **kwargs)

    def warning(self, message, *args, **kwargs):
        """
        Produces WARNING level log
        See :class:`logging.Logger`
        """
        self._logger.warning(message, *args, **kwargs)

    def debug(self, message, *args, **kwargs):
        """
        Produces DEBUG level log
        See :class:`logging.Logger`
        """
        self._logger.debug(message, *args, **kwargs)

    def trace(self, message, *args, **kwargs):
        """
        Produces TRACE level log
        See :class:`logging.Logger`
        """
        self._logger.trace(message, *args, **kwargs)

    def error(self, message, *args, **kwargs):
        """
        Produces ERROR level log
        See :class:`logging.Logger`
        """
        self._logger.error(message, *args, **kwargs)

    def critical(self, message, *args, **kwargs):
        """
        Produces CRITICAL level log
        See :class:`logging.Logger`
        """
        self._logger.critical(message, *args, **kwargs)


@decorator
def traced(obj, *args, **kwargs):
    """
    Decorator to attach to functions and methods to log at TRACE level.
    Trace level produced a log anytime a function or a method is entered and exited.
    """
    logger = logging.getLogger(generate_logger_name(obj))

    logger.trace("called")
    value = obj(*args, **kwargs)
    logger.trace("exited")

    return value


def generate_logger_name(obj):
    """
    Given a class method or a function it returns the logger name.

    Parameters
    ----------
    obj
        class method or function

    Returns
    -------
    str
        logger name
    """
    parent_logger_name = obj.__module__
    return "{}.{}".format(
        parent_logger_name, getattr(obj, "__qualname__", obj.__name__)
    )



import numpy as np

from paos import logger
from paos.core.coordinateBreak import coordinate_break


def raytrace(field, opt_chain, x=0.0, y=0.0):
    """
    Diagnostic function that implements the Paraxial ray-tracing and prints the output for each surface of the optical
    chain as the ray positions and slopes in the tangential and sagittal planes.

    Parameters
    ----------
    field: dict
        contains the slopes in the tangential and sagittal planes as field={'vt': slopey, 'vs': slopex}
    opt_chain: dict
        the dict of the optical elements returned by paos.parse_config
    x: float
        X-coordinate of the initial ray position
    y: float
        Y-coordinate of the initial ray position

    Returns
    -----
    out: list[str]
        A list of strings where each list item is the raytrace at a given surface.

    Examples
    --------

    >>> from paos.core.parseConfig import parse_config
    >>> from paos.core.raytrace import raytrace
    >>> pup_diameter, parameters, wavelengths, fields, opt_chains = parse_config('path/to/conf/file')
    >>> raytrace(fields[0], opt_chains[0])

    """
    vt = np.array([y, field["ut"]])
    vs = np.array([x, field["us"]])
    ostr = []
    for key, item in opt_chain.items():
        if item["type"] == "Coordinate Break":
            vt, vs = coordinate_break(
                vt,
                vs,
                item["xdec"],
                item["ydec"],
                item["xrot"],
                item["yrot"],
                0.0,
            )

        vt = item["ABCDt"]() @ vt
        vs = item["ABCDs"]() @ vs
        _ostr_ = "S{:02d} - {:15s} y:{:7.3f}mm ut:{:10.3e} rad x:{:7.3f}mm us:{:10.3e} rad".format(
            key, item["name"], 1000 * vt[0], vt[1], 1000 * vs[0], vs[1]
        )
        logger.debug(_ostr_)
        ostr.append(_ostr_)

    return ostr



import os
import sys

import numpy as np
import pandas as pd

from paos import logger
from paos.classes.abcd import ABCD
from paos.util.material import Material


def read_config(filename):
    """
    Given the input file name, it parses the simulation parameters and returns a dictionary.
    The input file is an Excel spreadsheet which contains three data sheets named 'general',
    'LD' and 'field'. 'general' contains the simulation wavelength, grid size and zoom, defined
    as the ratio of grid size to initial beam size in unit of pixel. 'LD' is the lens data and
    contains the sequence of surfaces for the simulation mimicking a Zemax lens data editor:
    supported surfaces include coordinate break, standard surface, obscuration, zernike,
    paraxial lens, slit and prism.

    Returns
    -------
    dict
        dictionary with the parsed input parameters for the simulation.

    Examples
    --------

    >>> from paos.paos_parseconfig_temp import read_config
    >>> simulation_parameters = read_config('path/to/conf/file')

    """

    parameters = {"general": {}, "LD": None, "field": {}}

    if not os.path.exists(filename) or not os.path.isfile(filename):
        logger.error(
            "Input file {} does not exist or is not a file. Quitting..".format(
                filename
            )
        )
        sys.exit()

    with pd.ExcelFile(filename, engine="openpyxl") as xls:
        wb = pd.read_excel(xls, "General")
        for key, val in zip(wb["INIT"], wb["Value"]):
            parameters["general"][key] = val
        wb = pd.read_excel(xls, "Lens Data")
        parameters["LD"] = wb

        wb = pd.read_excel(xls, "Fields")
        for field, x, y in zip(wb["Field"], wb["X"], wb["Y"]):
            parameters["field"][str(field)] = {
                "us": np.tan(np.deg2rad(x)),
                "ut": np.tan(np.deg2rad(y)),
            }

    return parameters


def ParseConfig(filename):
    """
    It parses the input file name and returns the input pupil diameter and three dictionaries for the simulation:
    one for the general parameters, one for the input fields and one for the optical chain.

    Returns
    -------
    dict
        the input pupil diameter, the general parameters, the input fields and the optical chain.

    Examples
    --------

    >>> from paos.paos_parseconfig_temp import ParseConfig
    >>> pupil_diameter, general, fields, optical_chain = ParseConfig('path/to/conf/file')

    """
    parameters = read_config(filename)

    wl = parameters["general"]["wavelength"]
    glasslib = Material(wl)

    n1 = None  # Refractive index
    pup_diameter = None  # input pupil pup_diameter

    opt_chain = {}

    for index, element in parameters["LD"].iterrows():

        chain_step = element["Surface num"]

        if element["Ignore"] == 1:
            continue

        if element["Surface Type"] == "INIT":
            n1 = 1.0  # propagation starts in free space
            xpup = element["XRADIUS"]
            ypup = element["YRADIUS"]

            if np.isfinite(xpup) and np.isfinite(ypup):
                pup_diameter = 2.0 * max(xpup, ypup)
            else:
                logger.error("Pupil wrongly defined")
                raise ValueError("Pupil wrongly defined")

            continue

        if n1 is None or pup_diameter is None:
            logger.error("INIT is not the first surface in Lens Data.")
            raise ValueError("INIT is not the first surface in Lens Data.")

        _data_ = {
            "num": element["Surface num"],
            "type": element["Surface Type"],
            "is_stop": True if element["Stop"] == 1 else False,
            "save": True if element["Save"] == 1 else False,
            "name": element["Comment"],
            "R": element["Radius"],
            "T": element["Thickness"],
            "material": element["Material"],
            "xrad": element["XRADIUS"],
            "yrad": element["YRADIUS"],
            "xdec": element["XDECENTER"],
            "ydec": element["YDECENTER"],
            "xrot": element["TiltAboutX"],
            "yrot": element["TiltAboutY"],
            "Mx": element["MagnificationX"],
            "My": element["MagnificationY"],
        }

        if element["Surface Type"] == "Zernike":
            sheet, stmp = element["Range"].split(".")
            colrange = "".join([i for i in stmp if not i.isdigit()])
            rowrange = "".join(
                [i for i in stmp if i.isdigit() or i == ":"]
            ).split(":")
            rowstart = int(rowrange[0])
            nrows = int(rowrange[1]) - rowstart + 1

            with pd.ExcelFile(filename, engine="openpyxl") as xls:
                wb0 = pd.read_excel(xls, sheet, header=None, nrows=3)
                wb1 = pd.read_excel(
                    xls,
                    sheet,
                    skiprows=rowstart - 1,
                    usecols="A," + colrange,
                    nrows=nrows,
                    header=None,
                )
            _data_.update(
                {
                    "Zindex": wb1[0].to_numpy(dtype=int),
                    "Z": wb1[1].to_numpy(dtype=float),
                    "Zradius": max(_data_["xrad"], _data_["yrad"]),
                    "Zwavelength": float(wb0[1][0]),
                    "Zordering": wb0[1][1].lower(),
                    "Znormalize": wb0[1][2],
                    "ABCDt": ABCD(),
                    "ABCDs": ABCD(),
                    "Zorigin": "x"  # this is the default origin for the angles: from x axis, counted
                    # positive counterclockwise
                }
            )

            opt_chain[chain_step] = _data_

        else:
            thickness = _data_["T"] if np.isfinite(_data_["T"]) else 0.0
            Mx = _data_["Mx"] if np.isfinite(_data_["Mx"]) else 1.0
            My = _data_["My"] if np.isfinite(_data_["My"]) else 1.0
            if _data_["type"] == "Coordinate Break":
                C = 0.0
                n2 = n1
            elif _data_["type"] == "Paraxial Lens":
                C = 1.0 / _data_["R"] if np.isfinite(_data_["R"]) else 0.0
                n2 = n1
            elif _data_["type"] in ("Standard", "Slit", "Obscuration"):
                C = 1.0 / _data_["R"] if np.isfinite(_data_["R"]) else 0.0
                if _data_["material"] == "MIRROR":
                    n2 = -n1
                elif _data_["material"] in glasslib.materials.keys():
                    n2 = glasslib.nmat(_data_["material"])[1] * np.sign(n1)
                else:
                    n2 = 1.0 * np.sign(n1)
            elif _data_["type"] == "Prism":
                C = 0.0
                n2 = n1
            else:
                logger.error(
                    "Surface Type not recognised: {:s}".format(
                        str(_data_["type"])
                    )
                )
                raise ValueError(
                    "Surface Type not recognised: {:s}".format(
                        str(_data_["type"])
                    )
                )

            _data_["ABCDt"] = ABCD(thickness, C, n1, n2, My)
            _data_["ABCDs"] = ABCD(thickness, C, n1, n2, Mx)

            logger.debug(
                "name: {}, curvature: {:4f}, thickness: {:.4f}, material: {}, n1: {:4f}, n2: {:4f}".format(
                    _data_["name"],
                    _data_["ABCDt"].power,
                    _data_["ABCDt"].thickness,
                    _data_["material"],
                    n1,
                    n2,
                )
            )

            n1 = n2
            opt_chain[chain_step] = _data_

    return pup_diameter, parameters["general"], parameters["field"], opt_chain
